""" Wire_Widget module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Optional, Dict, List, Union
import numpy as np
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QComboBox, QCheckBox
from magneticalc.QtWidgets2.QDoubleSpinBox2 import QDoubleSpinBox2
from magneticalc.QtWidgets2.QGroupBox2 import QGroupBox2
from magneticalc.QtWidgets2.QHLine import QHLine
from magneticalc.QtWidgets2.QIconLabel import QIconLabel
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.QtWidgets2.QPushButton2 import QPushButton2
from magneticalc.QtWidgets2.QSpinBox2 import QSpinBox2
from magneticalc.QTableView2 import QTableView2
from magneticalc.Debug import Debug
from magneticalc.ModelAccess import ModelAccess
from magneticalc.Theme import Theme


""" Default stretch setting. """
DefaultStretch = [1.0, 1.0, 1.0]


""" Default rotational symmetry setting. """
DefaultRotationalSymmetry = {
    "count" : 1,
    "radius": 0,
    "axis"  : 2,
    "offset": 0
}


class Wire_Widget(QGroupBox2):
    """ Wire_Widget class. """

    # Display settings
    UnitsLabelWidth = 26

    # Spinbox limits
    StretchMin = -99
    StretchMax = 99
    StretchStep = 0.1
    StretchPrecision = 2
    RotationalSymmetryCountMin = 1
    RotationalSymmetryCountMax = 99
    RotationalSymmetryRadiusMin = 0
    RotationalSymmetryRadiusMax = 99
    RotationalSymmetryRadiusStep = 0.1
    RotationalSymmetryRadiusPrecision = 2
    RotationalSymmetryOffsetMin = -360
    RotationalSymmetryOffsetMax = 360
    RotationalSymmetryOffsetStep = 1
    RotationalSymmetryOffsetPrecision = 1
    SlicerLimitMin = 0.001
    SlicerLimitMax = 2.0
    SlicerLimitStep = 0.001
    SlicerLimitPrecision = 3
    DcMin = -1e4
    DcMax = +1e4
    DcStep = 0.1
    DcPrecision = 3

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Populates the widget.

        @param gui: GUI
        """
        QGroupBox2.__init__(self, "Wire", color=Theme.DarkColor)
        Debug(self, ": Init", init=True)
        self.gui = gui

        # --------------------------------------------------------------------------------------------------------------

        table_icon_label = QIconLabel("Points", "mdi.vector-square", color=Theme.DarkColor, expand=False)
        table_icon_label.addWidget(QLabel2("⟨F2⟩, ⟨ESC⟩", font_size="13px", color=Theme.LiteColor, expand=False))
        table_icon_label.addStretch()
        table_icon_label.addWidget(QPushButton2("", "fa.plus", self.on_table_row_added))
        table_icon_label.addWidget(QLabel2("cm", align_right=True, width=self.UnitsLabelWidth))
        self.addLayout(table_icon_label)
        self.table = QTableView2(
            self.gui,
            row_count_minimum=2,
            col_keys=["X", "Y", "Z"],
            on_cell_edited=self.on_table_cell_edited,
            on_selection_changed=self.gui.redraw,
            on_row_deleted=self.on_table_row_deleted
        )
        self.addWidget(self.table)

        table_total_layout = QHBoxLayout()
        self.table_total_label = QLabel2("", color=Theme.MainColor, align_right=True)
        table_total_layout.addWidget(QLabel2("Total base points:", italic=True, color=Theme.LiteColor))
        table_total_layout.addWidget(self.table_total_label)
        self.addLayout(table_total_layout)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(QHLine())

        stretch_icon_label = QIconLabel("Stretch", "mdi.arrow-all", color=Theme.DarkColor)
        stretch_icon_label.addWidget(QPushButton2("", "fa.eraser", self.clear_stretch))
        stretch_icon_label.addWidget(QLabel2("cm", align_right=True, width=self.UnitsLabelWidth))
        self.addLayout(stretch_icon_label)

        self.stretch_spinbox = []
        stretch_layout = QHBoxLayout()
        for i in range(3):
            stretch_spinbox = QDoubleSpinBox2(
                gui=self.gui,
                minimum=self.StretchMin,
                maximum=self.StretchMax,
                step=self.StretchStep,
                precision=self.StretchPrecision,
                value=0,
                value_changed=self.set_stretch
            )
            self.stretch_spinbox.append(stretch_spinbox)
            stretch_layout.addWidget(QLabel2(["X", "Y", "Z"][i] + ":", expand=False))
            stretch_layout.addWidget(stretch_spinbox)
        self.addLayout(stretch_layout)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(QHLine())

        rotational_symmetry_icon_label = QIconLabel(
            "Rotational Symmetry", "mdi.rotate-3d-variant", color=Theme.DarkColor
        )
        rotational_symmetry_icon_label.addWidget(QPushButton2("", "fa.eraser", self.clear_rotational_symmetry))
        self.addLayout(rotational_symmetry_icon_label)
        rotational_symmetry_layout = QHBoxLayout()
        rotational_symmetry_layout_left = QVBoxLayout()
        rotational_symmetry_layout_middle = QVBoxLayout()
        rotational_symmetry_layout_right = QVBoxLayout()
        rotational_symmetry_layout.addLayout(rotational_symmetry_layout_left)
        rotational_symmetry_layout.addLayout(rotational_symmetry_layout_middle)
        rotational_symmetry_layout.addLayout(rotational_symmetry_layout_right)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        rotational_symmetry_layout_left.addWidget(QLabel2("Count:"))
        self.rotational_symmetry_count_spinbox = QSpinBox2(
            minimum=self.RotationalSymmetryCountMin,
            maximum=self.RotationalSymmetryCountMax,
            value=0,
            value_changed=self.set_rotational_symmetry
        )
        rotational_symmetry_layout_middle.addWidget(self.rotational_symmetry_count_spinbox)
        rotational_symmetry_layout_right.addWidget(QLabel2("", width=self.UnitsLabelWidth))

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.rotational_symmetry_radius_spinbox = QDoubleSpinBox2(
            gui=self.gui,
            minimum=self.RotationalSymmetryRadiusMin,
            maximum=self.RotationalSymmetryRadiusMax,
            step=self.RotationalSymmetryRadiusStep,
            precision=self.RotationalSymmetryRadiusPrecision,
            value=0,
            value_changed=self.set_rotational_symmetry
        )
        rotational_symmetry_layout_left.addWidget(QLabel2("Radius:"))
        rotational_symmetry_layout_middle.addWidget(self.rotational_symmetry_radius_spinbox)
        rotational_symmetry_count_radius_label = QLabel2("cm", align_right=True, width=self.UnitsLabelWidth)
        rotational_symmetry_layout_right.addWidget(rotational_symmetry_count_radius_label)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.rotational_symmetry_axis_combobox = QComboBox()
        for i, axis in enumerate(["X", "Y", "Z"]):
            self.rotational_symmetry_axis_combobox.addItem(axis)
        rotational_symmetry_layout_left.addWidget(QLabel2("Axis:"))
        rotational_symmetry_layout_middle.addWidget(self.rotational_symmetry_axis_combobox)
        rotational_symmetry_layout_right.addWidget(QLabel2("", width=self.UnitsLabelWidth))
        self.rotational_symmetry_axis_combobox.currentIndexChanged.connect(  # type: ignore
            self.set_rotational_symmetry
        )

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.rotational_symmetry_offset_spinbox = QDoubleSpinBox2(
            gui=self.gui,
            minimum=self.RotationalSymmetryOffsetMin,
            maximum=self.RotationalSymmetryOffsetMax,
            step=self.RotationalSymmetryOffsetStep,
            precision=self.RotationalSymmetryOffsetPrecision,
            value=0,
            value_changed=self.set_rotational_symmetry
        )
        rotational_symmetry_layout_left.addWidget(QLabel2("Offset Angle:"))
        rotational_symmetry_layout_middle.addWidget(self.rotational_symmetry_offset_spinbox)
        rotational_symmetry_offset_units_label = QLabel2("°", align_right=True, width=self.UnitsLabelWidth)
        rotational_symmetry_layout_right.addWidget(rotational_symmetry_offset_units_label)

        self.addLayout(rotational_symmetry_layout)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addWidget(QHLine())

        self.close_loop_checkbox = QCheckBox(" Close Loop")
        self.close_loop_checkbox.toggled.connect(  # type: ignore
            lambda: self.set_wire(_close_loop_=self.close_loop_checkbox.isChecked())
        )
        self.addWidget(self.close_loop_checkbox)

        rotational_symmetry_total_layout = QHBoxLayout()
        self.transformed_total_label = QLabel2("", color=Theme.MainColor, align_right=True)
        rotational_symmetry_total_layout.addWidget(
            QLabel2("Total transformed points:", italic=True, color=Theme.LiteColor)
        )
        rotational_symmetry_total_layout.addWidget(self.transformed_total_label)
        self.addLayout(rotational_symmetry_total_layout)
        self.addWidget(
            QPushButton2(
                " Replace base points",
                "mdi.content-copy",
                lambda: self.set_wire(_points_=self.gui.model.wire.points_transformed)
            )
        )

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(QHLine())

        self.addLayout(QIconLabel("Slicer Limit", "mdi.box-cutter", color=Theme.DarkColor))
        self.slicer_limit_spinbox = QDoubleSpinBox2(
            gui=self.gui,
            minimum=self.SlicerLimitMin,
            maximum=self.SlicerLimitMax,
            step=self.SlicerLimitStep,
            precision=self.SlicerLimitPrecision,
            value=0,
            value_changed=lambda: self.set_wire(_slicer_limit_=self.slicer_limit_spinbox.value())
        )
        slicer_limit_units_label = QLabel2("cm", align_right=True, width=self.UnitsLabelWidth)
        slicer_limit_layout = QHBoxLayout()
        slicer_limit_layout.addWidget(self.slicer_limit_spinbox)
        slicer_limit_layout.addWidget(slicer_limit_units_label)
        self.addLayout(slicer_limit_layout)

        sliced_total_layout = QHBoxLayout()
        self.sliced_total_label = QLabel2("", color=Theme.MainColor, align_right=True)
        sliced_total_layout.addWidget(QLabel2("Total sliced points:", italic=True, color=Theme.LiteColor))
        sliced_total_layout.addWidget(self.sliced_total_label)
        self.addLayout(sliced_total_layout)

        # --------------------------------------------------------------------------------------------------------------

        self.addWidget(QHLine())

        self.addLayout(QIconLabel("Wire Current", "fa.cog", color=Theme.DarkColor))
        self.dc_spinbox = QDoubleSpinBox2(
            gui=self.gui,
            minimum=self.DcMin,
            maximum=self.DcMax,
            step=self.DcStep,
            precision=self.DcPrecision,
            value=0,
            value_changed=lambda: self.set_wire(_dc_=self.dc_spinbox.value())
        )
        dc_layout = QHBoxLayout()
        dc_layout.addWidget(self.dc_spinbox)
        dc_layout.addWidget(QLabel2("A", align_right=True, width=self.UnitsLabelWidth))
        self.addLayout(dc_layout)

    def reload(self) -> None:
        """
        Reloads the widget.
        """
        Debug(self, ".reload()", refresh=True)

        self.blockSignals(True)

        wire_stretch = self.gui.project.get_point("wire_stretch")
        for i in range(3):
            self.stretch_spinbox[i].setValue(wire_stretch[i])

        self.rotational_symmetry_count_spinbox.setValue(self.gui.project.get_int("rotational_symmetry_count"))
        self.rotational_symmetry_radius_spinbox.setValue(self.gui.project.get_float("rotational_symmetry_radius"))

        rotational_symmetry_axis = self.gui.project.get_int("rotational_symmetry_axis")
        for i, axis in enumerate(["X", "Y", "Z"]):
            if i == rotational_symmetry_axis:
                self.rotational_symmetry_axis_combobox.setCurrentIndex(i)

        self.rotational_symmetry_offset_spinbox.setValue(self.gui.project.get_float("rotational_symmetry_offset"))

        self.close_loop_checkbox.setChecked(self.gui.project.get_bool("wire_close_loop"))

        self.slicer_limit_spinbox.setValue(self.gui.project.get_float("wire_slicer_limit"))
        self.dc_spinbox.setValue(self.gui.project.get_float("wire_dc"))

        self.blockSignals(False)

        # Initially load wire from project
        self.set_wire(recalculate=False, readjust_sampling_volume=False, invalidate=False)

        self.update()

    def update(self) -> None:
        """
        Updates the widget.
        """
        Debug(self, ".update()", refresh=True)

        self.update_labels()
        self.update_controls()

    def update_labels(self) -> None:
        """
        Updates the labels.
        """
        Debug(self, ".update_labels()", refresh=True)

        if self.gui.model.wire.valid:
            self.sliced_total_label.setText(str(len(self.gui.model.wire.points_sliced)))
        else:
            self.sliced_total_label.setText("N/A")

    def update_controls(self) -> None:
        """
        Updates the controls.
        """
        Debug(self, ".update_controls()", refresh=True)

        self.set_color(Theme.MainColor if self.gui.model.wire.valid else Theme.FailureColor)

    # ------------------------------------------------------------------------------------------------------------------

    def update_table(self) -> None:
        """
        Populates the table.
        """
        Debug(self, ".update_table()")

        points = [[f"{col:+0.02f}" for col in row] for row in self.gui.model.wire.points_base]

        self.table.set_data(
            data=points,
            row_keys=[str(i + 1) for i in range(len(self.gui.model.wire.points_base))]
        )
        self.table.select_last_row(focus=False)

        self.table_total_label.setText(str(len(self.gui.model.wire.points_base)))

    def on_table_row_added(self) -> None:
        """
        Gets called after a row has been added to the table.
        """
        Debug(self, ".on_table_row_added()")

        # Add a new base point (0, 0, 0) to the wire
        self.set_wire(_points_=list(self.gui.model.wire.points_base) + [np.zeros(3)])
        self.table.select_last_row()

    def on_table_cell_edited(self, value: float, row: int, column: int) -> None:
        """
        Gets called after a table cell has been edited.

        @param value: Cell value
        @param row: Row index
        @param column: Column index
        """
        points = self.gui.model.wire.points_base
        points[row][column] = value
        self.set_wire(_points_=points)

    def on_table_row_deleted(self, index: int) -> None:
        """
        Gets called after a row has been deleted from the table.

        @param index: Row index
        """
        Debug(self, ".on_table_row_deleted()")

        # Delete the wire base point at the given index
        self.set_wire(_points_=np.delete(self.gui.model.wire.points_base, index, axis=0))

    # ------------------------------------------------------------------------------------------------------------------

    def set_stretch(self) -> None:
        """
        Handles changes to stretch transform parameters.
        """
        if self.signalsBlocked():
            return

        Debug(self, ".set_stretch()")

        self.set_wire(_stretch_=[self.stretch_spinbox[i].value() for i in range(3)])

    def clear_stretch(self) -> None:
        """
        Clears the stretch values.
        """
        Debug(self, ".clear_stretch()")

        stretch = [1.0, 1.0, 1.0]
        self.set_wire(_stretch_=stretch)
        self.update_stretch(stretch=stretch)

    def update_stretch(self, stretch: List) -> None:
        """
        Updates the stretch controls.

        @param stretch: 3D point
        """
        Debug(self, ".update_stretch()")

        self.blockSignals(True)
        for i in range(3):
            self.stretch_spinbox[i].setValue(stretch[i])
        self.blockSignals(False)

    # ------------------------------------------------------------------------------------------------------------------

    def set_rotational_symmetry(self) -> None:
        """
        Handles changes to rotational symmetry transform parameters.
        """
        if self.signalsBlocked():
            return

        Debug(self, ".set_rotational_symmetry()")

        self.set_wire(
            _rotational_symmetry_={
                "count" : self.rotational_symmetry_count_spinbox.value(),
                "radius": self.rotational_symmetry_radius_spinbox.value(),
                "axis"  : {"X": 0, "Y": 1, "Z": 2}[self.rotational_symmetry_axis_combobox.currentText()],
                "offset": self.rotational_symmetry_offset_spinbox.value()
            }
        )

    def clear_rotational_symmetry(self) -> None:
        """
        Clears the rotational symmetry values.
        """
        Debug(self, ".clear_rotational_symmetry()")

        rotational_symmetry = {
            "count": 1,
            "radius": 0,
            "axis": 2,
            "offset": 0
        }
        self.set_wire(_rotational_symmetry_=rotational_symmetry)
        self.update_rotational_symmetry(rotational_symmetry=rotational_symmetry)

    def update_rotational_symmetry(self, rotational_symmetry: Dict) -> None:
        """
        Updates the rotational symmetry controls.

        @param rotational_symmetry: Dictionary
        """
        Debug(self, ".update_rotational_symmetry()")

        self.blockSignals(True)
        self.rotational_symmetry_count_spinbox.setValue(rotational_symmetry["count"])
        self.rotational_symmetry_radius_spinbox.setValue(rotational_symmetry["radius"])
        self.rotational_symmetry_axis_combobox.setCurrentIndex(rotational_symmetry["axis"])
        self.rotational_symmetry_offset_spinbox.setValue(rotational_symmetry["offset"])
        self.blockSignals(False)

    # ------------------------------------------------------------------------------------------------------------------

    def set_wire(
            self,
            _points_: Optional[Union[List, np.ndarray]] = None,
            _stretch_: Optional[List] = None,
            _rotational_symmetry_: Optional[Dict] = None,
            _close_loop_: Optional[bool] = None,
            _slicer_limit_: Optional[float] = None,
            _dc_: Optional[float] = None,
            invalidate: bool = True,
            recalculate: bool = True,
            readjust_sampling_volume: bool = True
    ) -> None:
        """
        Sets the wire. This will overwrite the currently set wire in the model.
        Any underscored parameter may be left set to None in order to load its default value.

        Note: Currently, only `stretch` and `rotational symmetry` controls are automatically updated by this function.

        @param _points_: Points (List of 3D points)
        @param _stretch_: XYZ stretch transform factors (3D point)
        @param _rotational_symmetry_: Dictionary for rotational symmetry transform
        @param _close_loop_: Enable to transform the wire into a closed loop (append first point)
        @param _slicer_limit_: Slicer limit
        @param _dc_: DC value
        @param invalidate: Enable to invalidate this model hierarchy level
        @param recalculate: Enable to trigger final recalculation
        @param readjust_sampling_volume: Enable to readjust sampling volume
        """
        if self.signalsBlocked():
            return

        Debug(self, ".set_wire()")

        with ModelAccess(self.gui, recalculate):

            should_update_stretch_controls = _stretch_ is not None
            should_update_rotational_symmetry_controls = _rotational_symmetry_ is not None

            points = self.gui.project.set_get_points("wire_points_base", _points_)
            stretch = self.gui.project.set_get_point("wire_stretch", _stretch_)
            close_loop = self.gui.project.set_get_bool("wire_close_loop", _close_loop_)
            slicer_limit = self.gui.project.set_get_float("wire_slicer_limit", _slicer_limit_)
            dc = self.gui.project.set_get_float("wire_dc", _dc_)

            rotational_symmetry = self.gui.project.set_get_dict(
                prefix="rotational_symmetry_",
                suffix="",
                types={"count": "int", "radius": "float", "axis": "int", "offset": "float"},
                values=_rotational_symmetry_,
            )

            self.gui.model.set_wire(
                invalidate=invalidate,
                points=points,
                stretch=stretch,
                rotational_symmetry=rotational_symmetry,
                close_loop=close_loop,
                slicer_limit=slicer_limit,
                dc=dc
            )

            self.update_table()

            if should_update_stretch_controls:
                self.update_stretch(stretch=stretch)

            if should_update_rotational_symmetry_controls:
                self.update_rotational_symmetry(rotational_symmetry=rotational_symmetry)

            self.transformed_total_label.setText(str(len(self.gui.model.wire.points_transformed)))

            if readjust_sampling_volume:
                self.gui.sidebar_left.sampling_volume_widget.readjust()



""" Constraint_Editor module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import List, Dict
from PyQt5.Qt import QShowEvent
from PyQt5.QtCore import Qt
from magneticalc.QtWidgets2.QDialog2 import QDialog2
from magneticalc.QtWidgets2.QIconLabel import QIconLabel
from magneticalc.QtWidgets2.QPushButton2 import QPushButton2
from magneticalc.QtWidgets2.QTextBrowser2 import QTextBrowser2
from magneticalc.Constraint import Constraint
from magneticalc.Comparison_Types import comparison_type_to_name, comparison_name_to_type
from magneticalc.Config_Group import Config_Collection
from magneticalc.Debug import Debug
from magneticalc.Norm_Types import norm_type_to_name, norm_name_to_type
from magneticalc.QTableView2 import QTableView2
from magneticalc.Theme import Theme


class Constraint_Editor(QDialog2):
    """ Constraint_Editor class. """

    # Constraint types
    Constraint_Types = {
        "norm"          : "str",
        "comparison"    : "str",
        "min"           : "float",
        "max"           : "float",
        "permeability"  : "float"
    }

    # Constraint options
    Constraint_Options = [
        [norm_type_to_name(norm_type) for norm_type in Constraint.Norm_Types_Supported],
        [comparison_type_to_name(comparison_type) for comparison_type in Constraint.Comparison_Types_Supported],
        None,
        None,
        None
    ]

    # HTML content
    HTML = f"""
        Lengths in <b>cm</b>. &nbsp; Angles in <b>degrees</b>.<br><br>

        <b style="color: {Theme.MainColor};">Experimental Feature</b><br><br>

        By default, all sampling volume points have relative permeability µ<sub>r</sub> = 1.<br>
        A constraint assigns some other µ<sub>r</sub> to some region of the sampling volume.<br>
        Setting the permeability of some region to zero locally disables the field calculation.<br>
        In case of ambiguous constraints, the ones with maximum permeability take precedence.<br><br>

        Constraints of identical permeability are effectively intersected (logical <code>AND</code> behaviour).<br>
        <i>Note:</i> You can achieve a logical <code>OR</code> behaviour using <i>De Morgan's laws</i>.<br>
        (<code>A</code> <b>Not</b> <i>In Range</i>)
        <code>AND</code>
        (<code>B</code> <b>Not</b> <i>In Range</i>) =
        <b>Not</b> (<code>A</code> <i>In Range</i> <code>OR</code> <code>B</code> <i>In Range</i>).<br>
        This result can then be inverted by letting <code>A</code> and <code>B</code> evaluate to µ<sub>r</sub> = 0.
    """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Prepares the constraint editor, but doesn't fully initialize it yet.

        @param gui: GUI
        """
        QDialog2.__init__(self, title="Constraint Editor", width=700)
        Debug(self, ": Init", init=True)
        self.gui = gui

        self._constraint_collection = Config_Collection(
            gui=gui,
            prefix="constraint_",
            types=self.Constraint_Types,
            first_without_suffix=False
        )

        table_icon_label = QIconLabel("Constraints", "mdi.playlist-edit", color=Theme.DarkColor)
        table_add_button = QPushButton2("Add constraint", "fa.plus", self.on_table_row_added)
        table_icon_label.addWidget(table_add_button)
        self.addLayout(table_icon_label)

        self.table = QTableView2(
            self.gui,
            row_prefix="constraint_",
            col_keys=["Norm", "Comparison", "Min.", "Max.", "µ (relative)"],
            col_types=self.Constraint_Types,
            col_options=self.Constraint_Options,
            on_cell_edited=self.on_table_cell_edited,
            on_row_deleted=self.on_table_row_deleted
        )
        self.addWidget(self.table)

        self.text_browser = QTextBrowser2(html=self.HTML)
        self.addWidget(self.text_browser, alignment=Qt.AlignBottom)

        self.addButtons({
            "OK": ("fa.check", self.accept)
        })

        self._changed = False

    def showEvent(self, event: QShowEvent) -> None:
        """
        Gets called when the dialog is opened.

        @param event: QShowEvent
        """
        Debug(self, ".showEvent()")
        self.text_browser.fit_to_contents()
        self.table.setFocus()

    @property
    def changed(self) -> bool:
        """
        Returns the "changed" state.
        """
        return self._changed

    @changed.setter
    def changed(self, changed: bool) -> None:
        """
        Sets the "changed" state.
        This updates the window title.
        """
        self._changed = changed
        self.setWindowTitle("Constraint Editor" + (" *" if changed else ""))

    def reload(self) -> None:
        """
        Reloads the constraint editor.
        """
        Debug(self, ".reload()", refresh=True)
        self.update_table()
        self.changed = False

    def update_table(self) -> None:
        """
        Populates the table.
        """
        self.table.set_data(
            data=[list(col_data.values()) for col_data in self.rows],
            row_keys=[str(i + 1) for i in range(self.rows_count)]
        )
        self.table.select_last_row(focus=False)

    # ------------------------------------------------------------------------------------------------------------------

    @property
    def rows(self) -> List[Dict]:
        """
        Gets the list of rows (constraint groups).

        @return: List of rows (constraint groups).
        """
        return self._constraint_collection.get_all_groups()

    @property
    def rows_count(self) -> int:
        """
        Gets the number of rows (constraint groups).

        @return: Number of rows (constraint groups)
        """
        return self._constraint_collection.get_count()

    def get_constraints(self) -> List[Constraint]:
        """
        Gets the list of constraints.

        @return: List of constraints
        """
        constraints = []
        for row in self.rows:
            Debug(self, f".get_constraints(): {row}")
            constraints.append(
                Constraint(
                    norm_name_to_type(row["norm"]),
                    comparison_name_to_type(row["comparison"]),
                    row["min"],
                    row["max"],
                    row["permeability"]
                )
            )
        return constraints

    # ------------------------------------------------------------------------------------------------------------------

    def on_table_row_added(self) -> None:
        """
        Gets called after a row has been added to the table.
        """
        self._constraint_collection.add_group(values={
            "norm"          : norm_type_to_name(Constraint.Norm_Types_Supported[0]),
            "comparison"    : comparison_type_to_name(Constraint.Comparison_Types_Supported[0]),
            "min"           : "0.0000",
            "max"           : "1.0000",
            "permeability"  : "1.0000"
        })

        self.update_table()
        self.table.select_last_row()
        self.changed = True

    def on_table_cell_edited(self, value: str, row: int, column: int):
        """
        Gets called after a table cell has been edited.

        @param value: Cell value
        @param row: Row index
        @param column: Column index
        """
        self._constraint_collection.set(group_index=row, key=list(self.Constraint_Types.keys())[column], value=value)
        self.changed = True

    def on_table_row_deleted(self, row: int) -> None:
        """
        Gets called after a row has been deleted from the table.

        @param row: Row index
        """
        self._constraint_collection.del_group(row)
        self.update_table()
        self.changed = True



""" QSaveAction module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
from datetime import datetime
from PyQt5.QtWidgets import QFileDialog


class QSaveAction:
    """ QSaveAction class. """

    def __init__(
            self,
            gui,
            title: str,
            date: bool,
            filename: str,
            extension: str,
            _filter: str
    ) -> None:
        """
        Initializes a save dialog.

        @param gui: GUI
        @param title: Title
        @param date: Enable datetime prefix
        @param filename: Filename
        @param extension: Extension
        @param _filter: Filename filters
        """
        self.filename, _chosen_extension = QFileDialog.getSaveFileName(
            parent=gui,
            caption=title,
            directory=(datetime.now().strftime("%Y-%m-%d_%H-%M-%S_") if date else "") + filename,
            filter=_filter,
            options=QFileDialog.DontUseNativeDialog
        )

        if self.filename == "":
            self.filename = None
            return

        _file_name, file_extension = os.path.splitext(self.filename)
        if file_extension.lower() != extension:
            self.filename += extension



""" Metric_Widget module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Optional, Dict
import numpy as np
from si_prefix import si_format
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QComboBox
from magneticalc.QtWidgets2.QGroupBox2 import QGroupBox2
from magneticalc.QtWidgets2.QHLine import QHLine
from magneticalc.QtWidgets2.QIconLabel import QIconLabel
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.Debug import Debug
from magneticalc.Metric_Presets import Metric_Presets
from magneticalc.ModelAccess import ModelAccess
from magneticalc.Theme import Theme


class Metric_Widget(QGroupBox2):
    """ Metric_Widget class. """

    # Formatting settings
    ValuePrecision = 1

    # Divergent metric gradient
    Cool_Gradient_CSS = """
        background: qlineargradient(
            x1:0 y1:0, x2:1 y2:0,
            stop:0 #00fffe,
            stop:1 #ff22f9
        );
    """

    # Cyclic metric gradient
    HSV_Gradient_CSS = """
        background: qlineargradient(
            x1:0 y1:0, x2:1 y2:0,
            stop:0.00 #ff0000,
            stop:0.08 #ff8000,
            stop:0.17 #ffff00,
            stop:0.25 #80ff00,
            stop:0.33 #00ff00,
            stop:0.42 #00ff80,
            stop:0.50 #00ffff,
            stop:0.58 #0080ff,
            stop:0.67 #0000ff,
            stop:0.75 #8000ff,
            stop:0.83 #ff00ff,
            stop:0.92 #ff0080,
            stop:1.00 #ff0000
        );
    """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Populates the widget.

        @param gui: GUI
        """
        QGroupBox2.__init__(self, "Metric", color=Theme.DarkColor)
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.addLayout(QIconLabel("Color", "fa.tint", color=Theme.DarkColor))
        self.color_metric_combobox = QComboBox()
        for i, preset in enumerate(Metric_Presets.List):
            self.color_metric_combobox.addItem(preset["id"])
        self.color_metric_combobox.currentIndexChanged.connect(  # type: ignore
            lambda: self.set_metric(_color_preset_=Metric_Presets.get_by_id(self.color_metric_combobox.currentText()))
        )
        self.addWidget(self.color_metric_combobox)

        self.color_metric_limits_layout = QHBoxLayout()
        self.color_metric_limits_widget = QWidget()
        self.color_metric_limits_widget.setLayout(self.color_metric_limits_layout)
        self.color_metric_min_label = QLabel2("N/A", css="background: none;", expand=False)
        self.color_metric_max_label = QLabel2("N/A", css="background: none;", expand=False)
        self.color_metric_limits_layout.addWidget(self.color_metric_min_label)
        self.color_metric_limits_layout.addStretch()
        self.color_metric_limits_layout.addWidget(
            QLabel2("", icon="mdi.dots-horizontal", icon_color=Theme.DarkColor, css="background: none;", expand=False)
        )
        self.color_metric_limits_layout.addStretch()
        self.color_metric_limits_layout.addWidget(self.color_metric_max_label)
        self.addWidget(self.color_metric_limits_widget)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addWidget(QHLine())

        self.addLayout(QIconLabel("Alpha", "mdi.blur", color=Theme.DarkColor))
        self.alpha_metric_combobox = QComboBox()
        for i, preset in enumerate(Metric_Presets.List):
            # Note: (v1.9)
            # Using angle metric for alpha transparency is discouraged and is not available in the combobox anymore.
            if not preset["is_angle"]:
                self.alpha_metric_combobox.addItem(preset["id"])
        self.alpha_metric_combobox.currentIndexChanged.connect(  # type: ignore
            lambda: self.set_metric(_alpha_preset_=Metric_Presets.get_by_id(self.alpha_metric_combobox.currentText()))
        )
        self.addWidget(self.alpha_metric_combobox)

        self.alpha_metric_limits_layout = QHBoxLayout()
        self.alpha_metric_limits_widget = QWidget()
        self.alpha_metric_limits_widget.setLayout(self.alpha_metric_limits_layout)
        self.alpha_metric_min_label = QLabel2("N/A", css="background: none;", expand=False)
        self.alpha_metric_max_label = QLabel2("N/A", css="background: none;", expand=False)
        self.alpha_metric_limits_layout.addWidget(self.alpha_metric_min_label)
        self.alpha_metric_limits_layout.addStretch()
        self.alpha_metric_limits_layout.addWidget(
            QLabel2("", icon="mdi.dots-horizontal", icon_color=Theme.DarkColor, css="background: none;", expand=False)
        )
        self.alpha_metric_limits_layout.addStretch()
        self.alpha_metric_limits_layout.addWidget(self.alpha_metric_max_label)
        self.addWidget(self.alpha_metric_limits_widget)

    def reload(self) -> None:
        """
        Reloads the widget.
        """
        Debug(self, ".reload()", refresh=True)

        self.blockSignals(True)

        color_metric = self.gui.project.get_str("color_metric")
        for i, preset in enumerate(Metric_Presets.List):
            if preset["id"] == color_metric:
                self.color_metric_combobox.setCurrentIndex(i)

        alpha_metric = self.gui.project.get_str("alpha_metric")
        for i, preset in enumerate(Metric_Presets.List):
            if preset["id"] == alpha_metric:
                self.alpha_metric_combobox.setCurrentIndex(i)

        self.blockSignals(False)

        # Initially load metric from project
        self.set_metric(recalculate=False, update_labels=False, invalidate=False)

        self.update()

    def update(self) -> None:
        """
        Updates the widget.
        """
        Debug(self, ".update()", refresh=True)

        self.update_labels()
        self.update_controls()

    def update_labels(self) -> None:
        """
        Updates the labels.
        """
        Debug(self, ".update_labels()", refresh=True)

        if self.gui.model.metric.valid:

            limits = self.gui.model.metric.limits

            show_gauss = self.gui.project.get_bool("show_gauss")
            field_units, field_factor = self.gui.model.field.get_units(show_gauss=show_gauss)

            if self.gui.model.metric.color_preset["is_angle"]:
                color_label_min = "0°"
                color_label_max = "360°"
                self.color_metric_limits_widget.setStyleSheet(self.HSV_Gradient_CSS)
                self.color_metric_min_label.setStyleSheet("background: none; color: #ffffff;")
                self.color_metric_max_label.setStyleSheet("background: none; color: #ffffff;")
            else:
                color_log_prefix, color_log_suffix = ("log(", ")") \
                    if self.gui.model.metric.color_preset["is_log"] else ("", "")

                if np.isnan(limits["color_min"]):
                    color_label_min = "NaN"
                else:
                    color_label_min = color_log_prefix +\
                        si_format(
                            limits["color_min"] * field_factor,
                            precision=self.ValuePrecision,
                            exp_format_str="{value}e{expof10} "
                        ) +\
                        field_units +\
                        color_log_suffix
                if np.isnan(limits["color_max"]):
                    color_label_max = "NaN"
                else:
                    color_label_max = color_log_prefix +\
                        si_format(
                            limits["color_max"] * field_factor,
                            precision=self.ValuePrecision,
                            exp_format_str="{value}e{expof10} "
                        ) +\
                        field_units + \
                        color_log_suffix
                self.color_metric_limits_widget.setStyleSheet(self.Cool_Gradient_CSS)
                self.color_metric_min_label.setStyleSheet("background: none; color: #000000; font-weight: bold;")
                self.color_metric_max_label.setStyleSheet("background: none; color: #ffffff; font-weight: bold;")

            if self.gui.model.metric.alpha_preset["is_angle"]:
                # Note: (v1.9)
                # Using angle metric for alpha transparency is discouraged and is not available in the combobox anymore.
                alpha_label_min = "0°"
                alpha_label_max = "360°"
                self.alpha_metric_limits_widget.setStyleSheet(self.HSV_Gradient_CSS)
                self.alpha_metric_min_label.setStyleSheet("background: none; color: #ffffff;")
                self.alpha_metric_max_label.setStyleSheet("background: none; color: #ffffff;")
            else:
                alpha_log_prefix, alpha_log_suffix = ("log(", ")") \
                    if self.gui.model.metric.alpha_preset["is_log"] else ("", "")

                if np.isnan(limits["alpha_min"]):
                    alpha_label_min = "NaN"
                else:
                    alpha_label_min = alpha_log_prefix +\
                        si_format(
                            limits["alpha_min"] * field_factor,
                            precision=self.ValuePrecision,
                            exp_format_str="{value}e{expof10} "
                        ) +\
                        field_units +\
                        alpha_log_suffix
                if np.isnan(limits["alpha_max"]):
                    alpha_label_max = "NaN"
                else:
                    alpha_label_max = alpha_log_prefix +\
                        si_format(
                            limits["alpha_max"] * field_factor,
                            precision=self.ValuePrecision,
                            exp_format_str="{value}e{expof10} "
                        ) +\
                        field_units + \
                        alpha_log_suffix
                self.alpha_metric_limits_widget.setStyleSheet(self.Cool_Gradient_CSS)
                self.alpha_metric_min_label.setStyleSheet("background: none; color: #000000; font-weight: bold;")
                self.alpha_metric_max_label.setStyleSheet("background: none; color: #ffffff; font-weight: bold;")

            self.color_metric_min_label.setText(color_label_min)
            self.color_metric_max_label.setText(color_label_max)
            self.alpha_metric_min_label.setText(alpha_label_min)
            self.alpha_metric_max_label.setText(alpha_label_max)

        else:

            self.color_metric_min_label.setText("N/A")
            self.color_metric_max_label.setText("N/A")
            self.color_metric_limits_widget.setStyleSheet("")
            self.color_metric_min_label.setStyleSheet("background: none; color: #000000;")
            self.color_metric_max_label.setStyleSheet("background: none; color: #000000;")

            self.alpha_metric_min_label.setText("N/A")
            self.alpha_metric_max_label.setText("N/A")
            self.alpha_metric_limits_widget.setStyleSheet("")
            self.alpha_metric_min_label.setStyleSheet("background: none; color: #000000;")
            self.alpha_metric_max_label.setStyleSheet("background: none; color: #000000;")

    def update_controls(self) -> None:
        """
        Updates the controls.
        """
        Debug(self, ".update_controls()", refresh=True)
        self.set_color(Theme.MainColor if self.gui.model.metric.valid else Theme.FailureColor)

    # ------------------------------------------------------------------------------------------------------------------

    def set_metric(
            self,
            _color_preset_: Optional[Dict] = None,
            _alpha_preset_: Optional[Dict] = None,
            invalidate: bool = True,
            recalculate: bool = True,
            update_labels: bool = True
    ) -> None:
        """
        Sets the metric. This will overwrite the currently set metric in the model.
        Any underscored parameter may be left set to None in order to load its default value.

        @param _color_preset_: Color metric preset (parameters, see Metric module)
        @param _alpha_preset_: Alpha metric preset (parameters, see Metric module)
        @param invalidate: Enable to invalidate this model hierarchy level
        @param recalculate: Enable to trigger final recalculation
        @param update_labels: Enable to update metric labels
        """
        if self.signalsBlocked():
            return

        Debug(self, ".set_metric()")

        with ModelAccess(self.gui, recalculate):

            if _color_preset_ is None:
                color_preset = Metric_Presets.get_by_id(self.gui.project.get_str("color_metric"))
            else:
                self.gui.project.set_str("color_metric", _color_preset_["id"])
                color_preset = _color_preset_

            if _alpha_preset_ is None:
                alpha_preset = Metric_Presets.get_by_id(self.gui.project.get_str("alpha_metric"))
            else:
                self.gui.project.set_str("alpha_metric", _alpha_preset_["id"])
                alpha_preset = _alpha_preset_

            self.gui.model.set_metric(
                invalidate=invalidate,
                color_preset=color_preset,
                alpha_preset=alpha_preset
            )

            if update_labels:
                self.update_labels()



""" CheckForUpdates_Dialog module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Optional, Tuple
import re
from urllib.error import URLError
from urllib.request import urlopen
from PyQt5.Qt import QFont, QSize
from PyQt5.QtWidgets import QTextEdit
from magneticalc.QtWidgets2.QDialog2 import QDialog2
from magneticalc.QtWidgets2.QIconLabel import QIconLabel
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.Debug import Debug
from magneticalc.Theme import Theme
from magneticalc.Version import __VERSION__, __PYPI_SIMPLE__URL__


class CheckForUpdates_Dialog(QDialog2):
    """ CheckForUpdates_Dialog class. """

    def __init__(self) -> None:
        """
        Initializes the dialog.
        """
        QDialog2.__init__(self, title="Check for Updates", width=500)
        Debug(self, ": Init", init=True)

        icon, string, success, error = self.check_for_updates()

        Debug(self, f": Check for Updates ({__PYPI_SIMPLE__URL__}): {string}", success=success, error=error)

        color = Theme.FailureColor if error else (Theme.SuccessColor if success else Theme.DialogTextColor)
        self.addLayout(
            QIconLabel(
                text=string,
                icon=icon,
                text_color=color,
                icon_color=color,
                icon_size=QSize(48, 48)
            )
        )

        if success:
            self.addSpacing(8)
            self.addWidget(QLabel2("Please update now:", italic=True))
            self.addSpacing(8)
            cmd = QTextEdit("python3 -m pip install magneticalc --upgrade")
            cmd.setFont(QFont(Theme.DefaultFontFace, 12))
            cmd.setMaximumHeight(64)
            cmd.setReadOnly(True)
            self.addWidget(cmd)

        self.addSpacing(8)

        self.addButtons({
            "Close": ("fa.check", self.accept)
        })

    def check_for_updates(self) -> Tuple[str, str, bool, bool]:
        """
        Checks for updates.

        @return: Icon, text, success flag, error flag
        """
        contents = self.remote_contents(url=__PYPI_SIMPLE__URL__, timeout=2)
        if contents is None:
            return "fa.exclamation-circle", "Network Error", False, True

        version = self.parse_contents(contents)
        if version is None:
            return "fa.exclamation-circle", "Invalid Format", False, True

        if version > __VERSION__:
            return "fa.info-circle", f"Newer version available ({version})", True, False
        elif version == __VERSION__:
            return "fa.check-circle", f"Up-to-date ({version})", False, False
        else:
            return "fa.exclamation-circle", f"Ahead of current release ({version})", False, True

    @staticmethod
    def remote_contents(url: str, timeout: int) -> Optional[str]:
        """
        Gets the contents of a remote file.

        @param url: URL
        @param timeout: Timeout (s)
        @return: Contents if successful, None otherwise
        """
        try:
            return urlopen(url=url, timeout=timeout).read().decode("utf-8")
        except URLError:
            return None

    @staticmethod
    def parse_contents(contents: str) -> Optional[str]:
        """
        Extracts and parses the version number from the raw Version.py contents.

        @param contents: Contents
        @return: Version string if successful, None otherwise
        """

        # noinspection RegExpAnonymousGroup
        pattern = re.compile(r"MagnetiCalc-(\d+)\.(\d+)\.(\d+)")

        result = pattern.findall(contents)

        if result is None:
            return None

        if len(result) == 0:
            return None

        return "v" + ".".join(result[-1])



""" QDialog2 module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Optional
from PyQt5.Qt import QShowEvent
from PyQt5.QtWidgets import QDialog
from magneticalc.QtWidgets2.QLayouted import QLayouted


class QDialog2(QDialog, QLayouted):
    """ QDialog2 class. """

    def __init__(
            self,
            title: Optional[str] = None,
            width: Optional[int] = None
    ) -> None:
        """
        Initializes QDialog2.

        @param title: Title
        @param width: Width
        """
        QDialog.__init__(self)
        QLayouted.__init__(self)
        self.install_layout(self)

        if title:
            self.setWindowTitle(title)

        if width:
            self.setMinimumWidth(width)

        self.user_accepted = None

    def show(self):
        """
        Shows this dialog.
        """
        self.user_accepted = self.exec() == QDialog.Accepted

    def showEvent(self, event: QShowEvent) -> None:
        """
        Gets called when the dialog is opened.

        @param event: QShowEvent
        """
        pass



""" MagnetiCalc_Data module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import numpy as np
from typing import Dict, Tuple, List
from collections.abc import MutableMapping


class MagnetiCalc_Data(MutableMapping):
    """ MagnetiCalc_Data class. """

    def __init__(self, data: Dict) -> None:
        """
        Initializes MagnetiCalc_Data class.

        This object can be accessed like a dictionary to get the data from an HDF5 export.

        Moreover, it provides convenience functions for accessing reshaped or transformed copies of this data.

        @param data: Dictionary
        """
        self.dictionary: Dict = {}
        self.update(data)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def get_wire(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Gets the wire points as three separate 1D arrays (raveled).

        @return: x, y, z
        """
        assert "wire_points" in self, "Sorry, there are no wire points in this data."
        wire_points = self["wire_points"]
        x, y, z = wire_points["x"], wire_points["y"], wire_points["z"]
        return x, y, z

    def get_wire_list(self) -> List:
        """
        Gets the wire points as a single list of 3D points (raveled).

        @return: List
        """
        return list(zip(self.get_wire()))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def get_current(self) -> float:
        """
        Gets the wire current.
        """
        assert "wire_current" in self, "Sorry, there is no wire current in this data."
        return self["wire_current"]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def get_dimension(self) -> Tuple[int, int, int]:
        """
        Gets the sampling volume dimension.

        @return: nx, ny, nz
        """
        fields = self._get_fields()
        return fields["nx"], fields["ny"], fields["nz"]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def get_axes(self, reduce: bool = False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Gets the sampling volume axes as three separate 1D arrays (raveled).

        @param reduce: Enable to reduce each raveled array to its minimal representation (axis ticks)
        @return: x, y, z
        """
        fields = self._get_fields()
        x, y, z = fields["x"], fields["y"], fields["z"]

        if reduce:
            x, y, z = np.array(sorted(list(set(x)))), np.array(sorted(list(set(y)))), np.array(sorted(list(set(z))))

        return x, y, z

    def get_axes_list(self, **kwargs) -> List:
        """
        Gets the sampling volume axes as a single list of 3D points (raveled).

        @return: List
        """
        return list(zip(self.get_axes(**kwargs)))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def get_field(self, field_type: str, as_3d: bool = False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Gets the field as three separate 1D arrays (raveled).

        @param field_type: Field type ("A" or "B")
        @param as_3d: Enable to transform each raveled 1D array into an unraveled 3D mesh (indexed by the minimal axes)
        """
        assert field_type in ["A", "B"], "Invalid field type"

        fields = self._get_fields()
        keys = [field_type + "_x", field_type + "_y", field_type + "_z"]
        assert all([key in fields for key in keys]), f"Sorry, there is no {field_type}-field in this data."
        field_x, field_y, field_z = [fields[key] for key in keys]

        if as_3d:
            shape_3d = self.get_dimension()
            field_x = np.reshape(field_x, shape_3d, order="F")
            field_y = np.reshape(field_y, shape_3d, order="F")
            field_z = np.reshape(field_z, shape_3d, order="F")

        return field_x, field_y, field_z

    def get_a_field(self, **kwargs) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Gets the A-field components as three separate 1D arrays (raveled).
        Please refer to L{get_field()} for possible arguments.

        @return: A_x, A_y, A_z
        """
        return self.get_field("A", **kwargs)

    def get_a_field_list(self, **kwargs) -> List:
        """
        Gets the A-field components as a single list of 3D points (raveled).
        Please refer to L{get_field()} for possible arguments.

        @return: List
        """
        return list(zip(self.get_a_field(**kwargs)))

    def get_b_field(self, **kwargs) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Gets the B-field components.
        Please refer to L{get_field()} for possible arguments.

        @return: B_x, B_y, B_z
        """
        return self.get_field("B", **kwargs)

    def get_b_field_list(self, **kwargs) -> List:
        """
        Gets the B-field components as a single list of 3D points (raveled).
        Please refer to L{get_field()} for possible arguments.

        @return: List
        """
        return list(zip(self.get_b_field(**kwargs)))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def _get_fields(self) -> Dict:
        """
        Gets the raw "fields" dictionary from the data.

        @return: Dictionary
        """
        assert "fields" in self, "Sorry, there are no fields in this data."
        return self["fields"]

    # ------------------------------------------------------------------------------------------------------------------

    def __getitem__(self, key):
        return self.dictionary[self._keytransform(key)]

    def __setitem__(self, key, value):
        self.dictionary[self._keytransform(key)] = value

    def __delitem__(self, key):
        del self.dictionary[self._keytransform(key)]

    def __iter__(self):
        return iter(self.dictionary)

    def __len__(self):
        return len(self.dictionary)

    # noinspection PyMethodMayBeStatic
    def _keytransform(self, key):
        return key



""" Assert_Dialog module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
from urllib.parse import urlencode, quote_plus
from magneticalc.QtWidgets2.QDialog2 import QDialog2
from magneticalc.QtWidgets2.QTextBrowser2 import QTextBrowser2
from magneticalc.Debug import Debug
from magneticalc.Theme import Theme
from magneticalc.Version import Version, __URL__


class Assert_Dialog:
    """ Assert_Dialog class. """

    def __init__(self, assertion: bool, message: str) -> None:
        """
        Shows a user dialog if an assertion failed. Intended for beta-testing.
        This allows the user to either quit or resume (possibly resulting in unstable behaviour).
        Furthermore, it provides a link for filing an issue on GitHub (partially prefilled).

        @param assertion: Boolean
        @param message: Error message
        """
        if assertion:
            return

        Debug(self, f": ERROR: Assertion failed: {message}", error=True)

        self._dialog = QDialog2(title="Assertion failed", width=600)

        # Generate Github issue URL
        issue_url = \
            f"{__URL__}/issues/new?" + \
            urlencode(
                {
                    "title": f"Assertion failed: {message}",
                    "labels": "bug",
                    "body":
                        f"**I discovered a bug in {Version.String}:**\n"
                        "```\n"
                        f"{message}\n"
                        "```\n\n"
                        f"**Steps to reproduce the problem:**\n\n"
                        "[_Please make sure to fill this in!_]"
                },
                quote_via=quote_plus
            )

        # HTML content
        html = f"""
            <span style="color: {Theme.MainColor};"><b>Sorry for the inconvenience!</b></span><br>
            <br>
            You seem to have discovered a bug in MagnetiCalc.<br>
            If this error persists, please
            <a href="{issue_url}">file an issue on GitHub</a>!<br>
            <br>
            <b>Error message:</b><br>
            <pre>{message}</pre>
            """

        text_browser = QTextBrowser2(html=html)
        self._dialog.addWidget(text_browser)

        buttons = self._dialog.addButtons({
            "Abort application"         : ("fa.times-circle", self._dialog.reject),
            "Resume (possibly unstable)": ("fa.play-circle", self._dialog.accept)
        })
        buttons[0].setFocus()

        self._dialog.show()

        if not self._dialog.user_accepted:
            sys.exit()



""" QPushButton2 module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Callable, Optional
import qtawesome as qta
from PyQt5.QtWidgets import QPushButton


class QPushButton2(QPushButton):
    """ QPushButton2 class. """

    def __init__(
            self,
            text: str,
            icon: str,
            clicked: Optional[Callable] = None,
            css: Optional[str] = None
    ) -> None:
        """
        Initializes a QPushButton.

        @param text: Text
        @param icon: Icon name
        @param clicked: Callback for button click
        @param css: Additional CSS (optional)
        """
        QPushButton.__init__(self)

        self.setText(text)

        if icon != "":
            self.setIcon(qta.icon(icon))

        if clicked is not None:
            self.clicked.connect(clicked)  # type: ignore

        if css is not None:
            self.setStyleSheet(css)



""" Metric module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from magneticalc.Norm_Types import *
from typing import Dict, List, Callable, Tuple
import numpy as np
from numba import jit, prange
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.Backend_Types import get_jit_enabled
from magneticalc.ConditionalDecorator import ConditionalDecorator
from magneticalc.Debug import Debug
from magneticalc.Validatable import Validatable, require_valid, validator


@ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=False)
def metric_norm(norm_type: int, vector: np.ndarray) -> float:
    """
    Calculates the selected norm of some vector.

    Note: For JIT to work, this must be declared at the top level.

    @param norm_type: Norm type
    @param vector: 3D vector
    @return: Scalar if successful, np.inf on error
    """
    if norm_type == NORM_TYPE_X:
        value = vector[0]
    elif norm_type == NORM_TYPE_Y:
        value = vector[1]
    elif norm_type == NORM_TYPE_Z:
        value = vector[2]
    elif norm_type == NORM_TYPE_RADIUS:
        value = np.sqrt(vector[0] ** 2 + vector[1] ** 2 + vector[2] ** 2)
    elif norm_type == NORM_TYPE_RADIUS_X:
        value = np.abs(vector[0])
    elif norm_type == NORM_TYPE_RADIUS_Y:
        value = np.abs(vector[1])
    elif norm_type == NORM_TYPE_RADIUS_Z:
        value = np.abs(vector[2])
    elif norm_type == NORM_TYPE_RADIUS_XY:
        value = np.sqrt(vector[0] ** 2 + vector[1] ** 2)
    elif norm_type == NORM_TYPE_RADIUS_XZ:
        value = np.sqrt(vector[1] ** 2 + vector[2] ** 2)
    elif norm_type == NORM_TYPE_RADIUS_YZ:
        value = np.sqrt(vector[2] ** 2 + vector[0] ** 2)
    elif norm_type == NORM_TYPE_ANGLE_XY:
        value = (np.arctan2(vector[0], vector[1]) + np.pi) / np.pi / 2
    elif norm_type == NORM_TYPE_ANGLE_XZ:
        value = (np.arctan2(vector[0], vector[2]) + np.pi) / np.pi / 2
    elif norm_type == NORM_TYPE_ANGLE_YZ:
        value = (np.arctan2(vector[1], vector[2]) + np.pi) / np.pi / 2
    elif norm_type == NORM_TYPE_DIVERGENCE:
        # Invalid norm ID for metric_norm(); divergence must be calculated using metric_divergence()
        value = np.inf
    else:
        # Invalid norm ID
        value = np.inf

    return value


@ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=False)
def metric_divergence(neighborhood_vectors: np.ndarray, dL: float, polarity: int) -> float:
    """
    Calculates the divergence of a sampling volume neighborhood.

    Note: For JIT to work, this must be declared at the top level.

    @param neighborhood_vectors: Sampling volume neighborhood vectors (six 3D vectors)
    @param dL: Length element
    @param polarity: Polarity filter (-1: Keep values <= 0; 0: Keep all values; +1: Keep values >= 0)
    """
    dxp = neighborhood_vectors[0][0]
    dxn = neighborhood_vectors[3][0]

    dyp = neighborhood_vectors[1][1]
    dyn = neighborhood_vectors[4][1]

    dzp = neighborhood_vectors[2][2]
    dzn = neighborhood_vectors[5][2]

    value = (dxp - dxn + dyp - dyn + dzp - dzn) / 2 / dL

    if polarity == -1:
        if value > 0:
            return np.NaN
        else:
            return -value   # Keep divergence positive, especially for use as alpha metric
    elif polarity == +1:
        if value < 0:
            return np.NaN
        else:
            return value
    else:
        return value


@ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=False)
def color_map_divergent(color_normalized: float) -> np.ndarray:
    """
    Maps normalized value to color, divergent.

    Note: For JIT to work, this must be declared at the top level.

    @param color_normalized: Normalized color value
    @return: R, G, B
    """
    rgb_0 = np.array([0x00, 0xff, 0xfe]) / 255
    rgb_1 = np.array([0xff, 0x22, 0xf9]) / 255
    return rgb_0 + (rgb_1 - rgb_0) * color_normalized


@ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=False)
def color_map_cyclic(color_normalized: float) -> np.ndarray:
    """
    Maps normalized value to color, cyclic.

    Note: For JIT to work, this must be declared at the top level.

    @param color_normalized: Normalized color value
    @return: R, G, B
    """
    hue = (color_normalized + 1 / 6) % 1
    x = hue * 6
    r = 1 if 0 <= x <= 1 or 5 <= x <= 6 else x - 4 if 4 <= x <= 5 else 2 - x if 1 <= x <= 2 else 0
    g = 1 if 1 <= x <= 3 else x if 0 <= x <= 1 else 4 - x if 3 <= x <= 4 else 0
    b = 1 if 3 <= x <= 5 else x - 2 if 2 <= x <= 3 else 6 - x if 5 <= x <= 6 else 0
    return np.array([r, g, b])


class Metric(Validatable):
    """
    Provides different metrics, used for mapping some field vector properties to some color and alpha range.
    """

    # Length scale
    LengthScale = 1e-2  # m  (1 cm)

    # Minimum argument limit for logarithmic scaling
    LogNormMin = 1e-12

    def __init__(self) -> None:
        """
        Initializes an empty metric.

        This class holds a pair of metric presets.
        Using these metric presets, colors (including alpha channel) and field limits are calculated.
        """
        Validatable.__init__(self)
        Debug(self, ": Init", init=True)

        # Parameters
        self.color_preset = {}
        self.alpha_preset = {}

        # Results
        self._colors: np.ndarray = np.array([])
        self._limits: Dict = {}

    def set(
            self,
            color_preset: Dict,
            alpha_preset: Dict
    ) -> None:
        """
        Sets the parameters.

        @param color_preset: Color metric preset (dictionary)
        @param alpha_preset: Alpha metric preset (dictionary)
        """
        self.color_preset = color_preset
        self.alpha_preset = alpha_preset

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    @ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=True)
    def _norm_worker(norm_type: int, vectors: np.ndarray) -> np.ndarray:
        """
        Calculates the norm values of a list of vectors.

        @param norm_type: Norm type
        @param vectors: Ordered list of 3D vectors
        @return: Norm values
        """
        values = np.zeros(len(vectors))

        for i in prange(len(vectors)):
            values[i] = metric_norm(norm_type, vectors[i])

        return values

    @staticmethod
    @ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=True)
    def _divergence_worker(
            sampling_volume_neighborhood_indices: List[np.ndarray],
            vectors: np.ndarray,
            dL: int,
            polarity: int
    ) -> np.ndarray:
        """
        Calculates the divergence values of a list of vectors.

        @param sampling_volume_neighborhood_indices: Ordered list of sampling volume neighborhood indices
        @param vectors: Ordered list of 3D vectors
        @param dL: Length element
        @param polarity: Polarity filter (-1: Keep values <= 0; 0: Keep all values; +1: Keep values >= 0)
        @return: Norm values
        """
        values = np.zeros(len(vectors))

        for i in prange(len(vectors)):

            neighbor_vectors = []
            for index in sampling_volume_neighborhood_indices[i]:
                if index != -1:
                    neighbor_vectors.append(vectors[index])
                else:
                    neighbor_vectors = None
                    break

            if neighbor_vectors is None:
                values[i] = np.nan  # At least one neighbor is out of bounds
            else:
                values[i] = metric_divergence(neighbor_vectors, dL, polarity)

        return values

    @staticmethod
    @ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=True)
    def _normalize_worker(
            color_map_id: str,
            color_is_log: bool,
            color_norm_values,
            color_norm_min: float,
            color_norm_max: float,
            alpha_is_log: bool,
            alpha_norm_values: np.ndarray,
            alpha_norm_min: float,
            alpha_norm_max: float,
            colors: np.ndarray
    ) -> np.ndarray:
        """
        Normalizes color and alpha norm values, populates final color values.

        @param color_map_id: Color map ID
        @param color_is_log: Selects logarithmic normalizer for color
        @param color_norm_values: Color norm values
        @param color_norm_min: Minimum color norm value
        @param color_norm_max: Maximum color norm value
        @param alpha_is_log: Selects logarithmic normalizer for alpha
        @param alpha_norm_values: Alpha norm values
        @param alpha_norm_min: Minimum alpha norm value
        @param alpha_norm_max: Maximum alpha norm value
        @param colors: Color value list to populate (list of 4D vectors)
        @return: Color values (list of 4D vectors)
        """

        # (1) Calculate color and alpha normalization
        # (2) Calculate colormap and assemble final color/alpha values
        #     Note: Of course, the alpha metric colormap is ignored

        for i in prange(len(color_norm_values)):

            color_normalized = color_norm_values[i]

            if np.isnan(color_normalized):
                # Replace NaN by minimum value
                color_normalized = color_norm_min
            else:
                # Clip before normalizing, needed for logarithmic normalization
                if color_normalized < color_norm_min:
                    color_normalized = color_norm_min
                elif color_normalized < color_norm_min:
                    color_normalized = color_norm_min

            color_normalized = (color_normalized - color_norm_min) / (color_norm_max - color_norm_min)
            if color_is_log:
                color_normalized = np.exp(color_normalized) / np.exp(1)

            if color_map_id == 0:
                # Divergent colormap
                rgb = color_map_divergent(color_normalized)
            else:
                # Cyclic colormap
                rgb = color_map_cyclic(color_normalized)

            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

            alpha_normalized = alpha_norm_values[i]

            if np.isnan(alpha_normalized):
                # Replace NaN by minimum value
                alpha_normalized = alpha_norm_min
            else:
                # Clip before normalizing, needed for logarithmic normalization
                if alpha_normalized < alpha_norm_min:
                    alpha_normalized = alpha_norm_min
                elif alpha_normalized > alpha_norm_max:
                    alpha_normalized = alpha_norm_max

            alpha_normalized = (alpha_normalized - alpha_norm_min) / (alpha_norm_max - alpha_norm_min)
            if alpha_is_log:
                alpha_normalized = np.exp(alpha_normalized) / np.exp(1)

            colors[i] = np.array([rgb[0], rgb[1], rgb[2], alpha_normalized])

        return colors

    @validator
    def recalculate(
            self,
            sampling_volume: SamplingVolume,  # type: ignore
            field: Field,  # type: ignore
            progress_callback: Callable
    ) -> bool:
        """
        Recalculates color and alpha values for field.

        @param sampling_volume: SamplingVolume
        @param field: Field
        @param progress_callback: Progress callback
        @return: True (currently non-interruptable)
        """
        Debug(self, ".recalculate()")

        n = field.vectors_count

        progress_callback(0)

        # Sampling volume length element (only needed for divergence metric)
        dL = Metric.LengthScale / sampling_volume.resolution

        # Calculate color metric values
        if self.color_preset["norm_type"] == NORM_TYPE_DIVERGENCE:
            # Calculate divergence
            color_norm_values = self._divergence_worker(
                sampling_volume.neighbor_indices,
                field.vectors,
                dL,
                self.color_preset["polarity"]
            )
        else:
            # Calculate other norm
            color_norm_values = self._norm_worker(
                self.color_preset["norm_type"],
                field.vectors
            )

        progress_callback(25)

        # Calculate alpha metric values
        if self.alpha_preset["norm_type"] == NORM_TYPE_DIVERGENCE:
            # Calculate divergence
            alpha_norm_values = self._divergence_worker(
                sampling_volume.neighbor_indices,
                field.vectors,
                dL,
                self.alpha_preset["polarity"]
            )
        else:
            # Calculate other norm
            alpha_norm_values = self._norm_worker(
                self.alpha_preset["norm_type"],
                field.vectors
            )

        progress_callback(50)

        # Select color range
        if self.color_preset["is_angle"]:
            color_norm_min, color_norm_max = 0.0, 1.0
        elif all(np.isnan(color_norm_values)):
            color_norm_min, color_norm_max = np.nan, np.nan
        else:
            color_norm_min, color_norm_max = np.nanmin(color_norm_values), np.nanmax(color_norm_values)

        # Select alpha range
        if self.alpha_preset["is_angle"]:
            alpha_norm_min, alpha_norm_max = 0.0, 1.0
        elif all(np.isnan(alpha_norm_values)):
            alpha_norm_min, alpha_norm_max = np.nan, np.nan
        else:
            alpha_norm_min, alpha_norm_max = np.nanmin(alpha_norm_values), np.nanmax(alpha_norm_values)

        # Select color normalizer
        if self.color_preset["is_log"]:
            # Logarithmic normalization

            Assert_Dialog(not self.color_preset["is_angle"], "Logarithmic angles don't make any sense")

            # Adjust range for logarithm (out-of-range values will be clipped in the loop below)
            color_norm_min_ = max(color_norm_min, Metric.LogNormMin)    # avoiding ValueError: min must be positive
            color_norm_max_ = max(color_norm_min_, color_norm_max)          # avoiding ValueError: min must be <= max
        else:
            # Linear normalization
            color_norm_min_ = color_norm_min
            color_norm_max_ = color_norm_max

        # Select alpha normalizer
        if self.alpha_preset["is_log"]:
            # Logarithmic normalization

            Assert_Dialog(not self.alpha_preset["is_angle"], "Logarithmic angles don't make any sense")

            # Adjust range for logarithm (out-of-range values will be clipped in the loop below)
            alpha_norm_min_ = max(alpha_norm_min, Metric.LogNormMin)    # avoiding ValueError: min must be positive
            alpha_norm_max_ = max(alpha_norm_min_, alpha_norm_max)          # avoiding ValueError: min must be <= max
        else:
            # Linear normalization
            alpha_norm_min_ = alpha_norm_min
            alpha_norm_max_ = alpha_norm_max

        colors = np.zeros([n, 4])

        # Calculate final color values
        # Note: If using logarithmic scaling, out-of-range values (still exceeding [0…1] here) may be clipped
        colors = self._normalize_worker(
            self.color_preset["colormap"],
            self.color_preset["is_log"],
            color_norm_values,
            color_norm_min_,
            color_norm_max_,
            self.alpha_preset["is_log"],
            alpha_norm_values,
            alpha_norm_min_,
            alpha_norm_max_,
            colors
        )

        self._colors = colors

        self._limits = {
            "color_min": color_norm_min,
            "color_max": color_norm_max,
            "alpha_min": alpha_norm_min,
            "alpha_max": alpha_norm_max
        }

        progress_callback(100)

        return True

    @property
    @require_valid
    def colors(self) -> np.ndarray:
        """
        Returns calculated colors.

        @return: Ordered list of color values (4-tuples)
        """
        return self._colors

    @property
    @require_valid
    def limits(self) -> Dict:
        """
        Returns calculated limits.

        @return: Dictionary
        """
        return self._limits

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    @ConditionalDecorator(get_jit_enabled(), jit, nopython=True, parallel=True)
    def boost_colors(boost: float, direction: float, colors: np.ndarray) -> np.ndarray:
        """
        "Boosts" an array of color values.

        @param boost: Boost value
        @param direction: Boost direction
        @param colors: Colors (ordered list of 4-tuples)
        @return: Colors (ordered list of 4-tuples)
        """
        for i in prange(len(colors)):
            r = np.max(np.array([0.0, np.min(np.array([1.0, colors[i][0] + boost * direction]))]))
            g = np.max(np.array([0.0, np.min(np.array([1.0, colors[i][1] + boost * direction]))]))
            b = np.max(np.array([0.0, np.min(np.array([1.0, colors[i][2] + boost * direction]))]))
            a = np.max(np.array([0.0, np.min(np.array([1.0, colors[i][3] + boost]))]))
            colors[i] = np.array([r, g, b, a])

        return colors



""" GUI module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import time
import atexit
from sty import fg
import qtawesome as qta
from PyQt5.Qt import QCloseEvent, QKeyEvent, QTimer
from PyQt5.QtCore import Qt, QThread, pyqtSignal, QLocale
from PyQt5.QtWidgets import QMainWindow, QSplitter
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.CalculationThread import CalculationThread
from magneticalc.Debug import Debug
from magneticalc.Menu import Menu
from magneticalc.Model import Model
from magneticalc.Project import Project
from magneticalc.SidebarLeft import SidebarLeft
from magneticalc.SidebarRight import SidebarRight
from magneticalc.Statusbar import Statusbar
from magneticalc.Theme import Theme
from magneticalc.VisPyCanvas import VisPyCanvas


class GUI(QMainWindow):
    """ GUI class. """

    # Used by L{Debug}
    DebugColor = fg.blue

    # Minimum window size
    MinimumWindowSize = (800, 600)

    # These signals are fired from the calculation thread
    calculation_status = pyqtSignal(str)
    calculation_exited = pyqtSignal(bool)

    # Used by ModelAccess to invalidate the statusbar
    invalidate_statusbar = pyqtSignal()

    def __init__(self) -> None:
        """
        Initializes the GUI.
        """
        QMainWindow.__init__(self, flags=Qt.Window)
        Debug(self, ": Init", init=True)

        self.initializing = True
        self.exiting = False

        self.user_locale = QLocale(QLocale.English)

        self.setWindowIcon(qta.icon("ei.magnet", color=Theme.MainColor))
        self.setMinimumSize(*self.MinimumWindowSize)
        self.showMaximized()

        self.project = Project(self)
        self.project.open_default(reload=False)

        # The calculation thread is started once initially; after that, recalculation is triggered through ModelAccess
        self.calculation_thread = None  # Will be initialized by "recalculate()" but is needed here for ModelAccess
        self.calculation_start_time = time.monotonic()  # Will be overwritten by "recalculate()"

        # Register exit handler (used by Assert_Dialog to exit gracefully)
        atexit.register(self.cleanup)

        # Create the statusbar first, as it connects to the "invalidate_statusbar" signal emitted by ModelAccess
        self.statusbar = Statusbar(self)
        self.invalidate_statusbar.connect(self.statusbar.invalidate)  # type: ignore

        # Create the model next, as the following objects will access it (each widget acts as view *and* controller)
        self.model = Model(self)

        # Create the left and right sidebar
        # Note: These create the wire, sampling volume, field, metric and parameters widgets,
        #       each populating the model from the project
        self.sidebar_left = SidebarLeft(self)
        self.sidebar_right = SidebarRight(self)

        # Create the VisPy canvas (our 3D scene)
        self.vispy_canvas = VisPyCanvas(self)
        self.vispy_canvas.native.setFocusPolicy(Qt.NoFocus)  # Don't let VisPy gain control – handle all events in GUI

        # Insert left sidebar, VisPy canvas and right sidebar into main layout.
        self.splitter = QSplitter(Qt.Horizontal)
        self.splitter.addWidget(self.sidebar_left)
        self.splitter.addWidget(self.vispy_canvas.native)
        self.splitter.addWidget(self.sidebar_right)
        self.setCentralWidget(self.splitter)
        self.splitter.setHandleWidth(8)

        # Create the menu
        self.menu = Menu(self)

        self.reload()

    # ------------------------------------------------------------------------------------------------------------------

    def reload(self) -> None:
        """
        Reloads the GUI.
        """
        self.sidebar_left.wire_widget.reload()
        self.sidebar_left.sampling_volume_widget.reload()
        self.sidebar_right.field_widget.reload()
        self.sidebar_right.metric_widget.reload()
        self.sidebar_right.parameters_widget.reload()
        self.sidebar_right.perspective_widget.reload()
        self.sidebar_right.display_widget.reload()

        self.menu.reload()
        self.statusbar.reload()

        self.vispy_canvas.load_perspective()

        Debug(self, ".reload(): Waiting 1 second for GUI update")
        QTimer.singleShot(1000, self.re_wrapper)

    def re_wrapper(self):
        """
        Simple wrapper that decides between recalculation or redraw.
        """
        if self.project.get_bool("auto_calculation"):
            self.recalculate()
        else:
            self.redraw()

    def redraw(self) -> None:
        """
        Redraws the scene.
        """
        if self.calculation_thread is not None:
            if self.calculation_thread.isRunning():
                Debug(self, ".redraw(): WARNING: Skipped because calculation is in progress", warning=True)
                return
            else:
                Debug(self, ".redraw(): WARNING: Setting calculation thread to None", warning=True)
                self.calculation_thread = None
        else:
            Debug(self, ".redraw()")

        self.sidebar_right.display_widget.setEnabled(self.model.field.valid)

        self.vispy_canvas.redraw()

    def recalculate(self) -> None:
        """
        Recalculates the model.
        """
        Debug(self, ".recalculate()")

        if self.calculation_thread is not None:
            Debug(self, ".recalculate(): WARNING: Killing orphaned calculation thread", warning=True)
            self.interrupt_calculation()

        # (Re-)connect all calculation thread communication signals.
        if not self.initializing:
            self.calculation_status.disconnect(self.statusbar.set_text)  # type: ignore
            self.calculation_exited.disconnect(self.on_calculation_exited)  # type: ignore
        self.calculation_status.connect(self.statusbar.set_text)  # type: ignore
        self.calculation_exited.connect(self.on_calculation_exited)  # type: ignore

        if self.initializing:
            self.initializing = False
            self.vispy_canvas.initializing = True

        self.redraw()
        self.statusbar.arm()

        # Create a new calculation thread and kick it off
        self.calculation_thread = CalculationThread(self)
        self.calculation_start_time = time.monotonic()
        self.calculation_thread.start()

    def interrupt_calculation(self) -> None:
        """
        Kills any running calculation.
        """
        if self.calculation_thread is None:
            Debug(self, ".interrupt_calculation(): WARNING: No calculation thread to interrupt", warning=True)
            return

        self.calculation_thread.disconnect_signals()

        if self.calculation_thread.isRunning():
            Debug(self, ".interrupt_calculation(): WARNING: Requesting interruption", warning=True)
            self.calculation_thread.requestInterruption()

            if self.calculation_thread.wait(5000):
                Debug(self, ".interrupt_calculation(): Exited gracefully", success=True)
            else:
                Assert_Dialog(False, "Failed to terminate calculation thread")
                if self.calculation_thread is not None:
                    if self.calculation_thread.isRunning():
                        Debug(self, ".interrupt_calculation(): WARNING: Terminating ungracefully", warning=True)
                        self.calculation_thread.terminate()
                        self.calculation_thread.wait()
        else:
            Debug(self, ".interrupt_calculation(): WARNING: Calculation thread should be running", warning=True)

        self.calculation_thread = None

    def on_calculation_exited(self, success: bool) -> None:
        """
        This is called after calculation thread has exited.

        @param success: True if calculation was successful, False otherwise
        """
        # Ignore when exiting
        if self.exiting:
            return

        calculation_time = time.monotonic() - self.calculation_start_time

        if self.calculation_thread is not None:
            if self.calculation_thread.isRunning():
                # Skipping because another thread is now running
                # Note: This happens all the time when calculation is interrupted and restarted through ModelAccess;
                #       we see this because there is no reliable way to revoke the delayed "calculation_exited" signal
                #       after another thread has already been started
                return
            else:
                # This happens when calculation finished and no other thread was started
                self.calculation_thread = None
                Debug(
                    self,
                    f".on_calculation_exited(success={success}): "
                    f"Took {calculation_time:.2f} s",
                    success=success
                )
        else:
            Debug(
                self,
                f".on_calculation_exited(success={success}): "
                f"WARNING: Interrupted after {calculation_time:.2f} s",
                warning=True
            )

        # Note: For some reason, most of the time we need an additional ("final-final") redraw here; VisPy glitch?
        self.redraw()

        self.statusbar.disarm(success)

    # ------------------------------------------------------------------------------------------------------------------

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """
        Handles key press event.

        @param event: Key press event
        """
        if event.key() == Qt.Key_F2:

            # Focus the  wire base points table
            self.sidebar_left.wire_widget.table.setFocus()

        elif event.key() == Qt.Key_F3:

            # Open the constraint editor
            self.sidebar_left.sampling_volume_widget.open_constraint_editor()

        elif event.key() == Qt.Key_F5:

            # Focus the main window (make sure to un-focus the wire base points table)
            self.setFocus()

            # (Re)start calculation
            self.recalculate()

        elif event.key() == Qt.Key_Escape:

            if self.sidebar_left.wire_widget.table.hasFocus():
                # Focus the main window, thus un-focusing the wire base points table
                self.setFocus()
            else:
                # Stop any running calculation
                if self.calculation_thread is not None:
                    if self.calculation_thread.isRunning():
                        # Cancel the running calculation
                        self.interrupt_calculation()

    # ------------------------------------------------------------------------------------------------------------------

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Gets called when the window is closed (programmatically through close() or manually)

        @param event: Close event
        """
        Debug(self, ".closeEvent()")

        closed = self.project.close(invalidate=False)
        if not closed:
            event.ignore()
            return

        self.cleanup()

    def cleanup(self):
        """
        Perform clean-up upon quitting the application.
        """
        Debug(self, ".cleanup()")

        # Unregister exit handler (used by Assert_Dialog to exit gracefully)
        atexit.unregister(self.cleanup)

        self.exiting = True

        if self.calculation_thread != QThread.currentThread():
            if self.calculation_thread is not None:
                self.interrupt_calculation()
        else:
            Debug(self, ".quit(): Called from calculation thread (assertion failed)")

        print()
        print("Goodbye!")



""" Debug module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Optional
from inspect import isclass, stack
from sty import Style, fg, ef


# Enable to see JIT debug output:
# import os
# os.environ["NUMBA_PARALLEL_DIAGNOSTICS"] = "4"


class Debug:
    """ Debug class. """

    # Colors
    LightColor = fg.grey
    SuccessColor = fg.green
    WarningColor = fg.magenta
    ErrorColor = fg.red

    # Enable debug output where init=True
    EnableInit = False

    # Enable debug output where refresh=True
    EnableRefresh = False

    # Block debug output from specific classes
    EnableBlacklist = True
    Blacklist = [
        "About_Dialog",
        "CheckForUpdates_Dialog",
        "Constraint_Editor",
        "Display_Widget",
        "ExportContainer_Dialog",
        "Field_Widget",
        "Metric_Widget",
        "ModelAccess",
        "OverridePadding_Dialog",
        "Parameters_Widget",
        "Perspective_Widget",
        "QTableView2",
        "SamplingVolume_Widget",
        "Usage_Dialog",
        "VisPyCanvas",
        "Wire_Widget",
    ]

    def __init__(
            self,
            obj: object,
            text: str,
            color: Optional[Style] = None,
            force: bool = False,
            success: bool = False,
            warning: bool = False,
            error: bool = False,
            init: bool = False,
            refresh: bool = False
    ) -> None:
        """
        Displays a colorful debug message and the current call hierarchy.

        @param obj: Class instance
        @param text: Debug message
        @param color: Color (may be None)
        @param force: Enable to override whitelist
        @param success: Enable to set color=SuccessColor
        @param warning: Enable to set color=WarningColor and force=True
        @param error: Enable to set color=ErrorColor and force=True
        @param init: Enable to mark as init message (filtering through EnableInit)
        @param refresh: Enable to mark as refresh message (filtering through EnableRefresh)
        """
        if init and not self.EnableInit:
            return

        if refresh and not self.EnableRefresh:
            return

        if isclass(obj):
            # Called from within class method, i.e. Debug(self, …)
            name = obj.__name__
        else:
            # Called from within instance / static method
            name = type(obj).__name__

        if warning or error:
            force = True

        if not force:
            # Skip blacklisted class names
            if self.EnableBlacklist and name in self.Blacklist:
                return

        # A class may specify its own default color
        if color is None:
            if hasattr(obj, "DebugColor"):
                color = obj.DebugColor  # type: ignore
            else:
                color = ""

        # Format call hierarchy
        hierarchy = ""
        for f in reversed(stack()[2:]):
            func_str = str(f.function)
            if func_str in ["<module>", "_run_module_as_main", "_run_code"]:
                # Called from within the main module (don't display these elements)
                continue
            if func_str == "run":
                # Called from within another thread (it is the calculation thread)
                func_str = "\tCalculationThread"
            hierarchy += func_str + "/"

        if success:
            color = self.SuccessColor

        if warning:
            color = self.WarningColor

        if error:
            color = self.ErrorColor

        print(self.LightColor + hierarchy + fg.rs + color + ef.bold + name + ef.rs + text + fg.rs + "\n", end="")



""" Menu module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from magneticalc.Debug import Debug
from magneticalc.File_Menu import File_Menu
from magneticalc.Help_Menu import Help_Menu
from magneticalc.Options_Menu import Options_Menu
from magneticalc.View_Menu import View_Menu
from magneticalc.Wire_Menu import Wire_Menu


class Menu:
    """ Menu class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Creates the menu bar.

        @param gui: GUI
        """
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.file_menu = File_Menu(self.gui)
        self.wire_menu = Wire_Menu(self.gui)
        self.view_menu = View_Menu(self.gui)
        self.options_menu = Options_Menu(self.gui)
        self.help_menu = Help_Menu(self.gui)

        self.gui.menuBar().addMenu(self.file_menu)
        self.gui.menuBar().addMenu(self.wire_menu)
        self.gui.menuBar().addMenu(self.view_menu)
        self.gui.menuBar().addMenu(self.options_menu)
        self.gui.menuBar().addMenu(self.help_menu)

    def reload(self) -> None:
        """
        Reloads the menu.
        """
        Debug(self, ".reload()", refresh=True)
        self.update()

    def update(self) -> None:
        """
        Updates all menus.
        """
        Debug(self, ".update()", refresh=True)
        self.file_menu.update()
        self.wire_menu.update()
        self.view_menu.update()
        self.options_menu.update()



""" Options_Menu module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from functools import partial
from PyQt5.QtWidgets import QMenu, QAction, QActionGroup
from magneticalc.Backend_Types import Backend_Types_Available, backend_type_to_name
from magneticalc.Debug import Debug


class Options_Menu(QMenu):
    """ Options_Menu class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ):
        QMenu.__init__(self, "&Options")
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.options_backend_group = QActionGroup(self.gui)
        self.options_backend_group.setExclusive(True)

        self.backend_actions = {}
        for backend_type, backend_available in Backend_Types_Available.items():
            self.backend_actions[backend_type] = QAction("Backend: " + backend_type_to_name(backend_type))
            self.backend_actions[backend_type].setEnabled(backend_available)
            self.backend_actions[backend_type].setCheckable(True)
            self.options_backend_group.addAction(self.backend_actions[backend_type])
            self.addAction(self.backend_actions[backend_type])
        for backend_type in Backend_Types_Available:
            self.backend_actions[backend_type].changed.connect(  # type: ignore
                partial(self.on_backend_changed, backend_type)
            )

    def update(self):
        """
        Updates the menu.
        """
        Debug(self, ".update()", refresh=True)
        self.gui.blockSignals(True)

        backend_type = self.gui.project.get_int("backend_type")
        for _backend_type in Backend_Types_Available:
            self.backend_actions[_backend_type].setChecked(backend_type == _backend_type)

        self.gui.blockSignals(False)

    # ------------------------------------------------------------------------------------------------------------------

    def on_backend_changed(self, backend_type: int) -> None:
        """
        Gets called when the backend changed.

        @param backend_type: Backend type
        """
        if self.gui.signalsBlocked():
            return

        if self.backend_actions[backend_type].isChecked():
            self.gui.project.set_int("backend_type", backend_type)
            self.gui.sidebar_right.field_widget.set_field()



""" Wire_Menu module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from functools import partial
import qtawesome as qta
from PyQt5.QtWidgets import QMenu, QFileDialog, QAction
from magneticalc.QtWidgets2.QSaveAction import QSaveAction
from magneticalc import API
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.Debug import Debug
from magneticalc.Wire_Widget import DefaultStretch, DefaultRotationalSymmetry
from magneticalc.Wire_Presets import Wire_Presets


class Wire_Menu(QMenu):
    """ Wire_Menu class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ):
        """
        Initializes the menu.
        """
        QMenu.__init__(self, "&Wire")
        Debug(self, ": Init", init=True)
        self.gui = gui

        load_preset_menu = QMenu("Load &Preset", self.gui)
        load_preset_menu.setIcon(qta.icon("mdi.vector-square"))
        for preset in Wire_Presets.List:
            action = QAction(qta.icon("mdi.vector-square"), preset["id"], self)
            action.triggered.connect(  # type: ignore
                partial(
                    self.gui.sidebar_left.wire_widget.set_wire,
                    _points_=preset["points"], _stretch_=DefaultStretch, _rotational_symmetry_=DefaultRotationalSymmetry
                )
            )
            load_preset_menu.addAction(action)
        self.addMenu(load_preset_menu)

        self.addSeparator()

        self.import_wire_action = QAction(qta.icon("fa.folder"), "&Import TXT …")
        self.import_wire_action.triggered.connect(  # type: ignore
            self.wire_import
        )
        self.addAction(self.import_wire_action)

        self.export_wire_action = QAction(qta.icon("fa.save"), "&Export TXT …")
        self.export_wire_action.triggered.connect(  # type: ignore
            self.wire_export
        )
        self.addAction(self.export_wire_action)

        self.update()

    def update(self):
        """
        Updates the menu.
        """
        Debug(self, ".update()", refresh=True)
        self.export_wire_action.setEnabled(self.gui.model.wire.valid)

    # ------------------------------------------------------------------------------------------------------------------

    def wire_import(self) -> None:
        """
        Imports wire points from a TXT file.
        """
        Debug(self.gui, ".wire_import()")

        filename, _chosen_extension = QFileDialog.getOpenFileName(
            parent=self.gui,
            caption="Import Wire",
            filter="Text File (*.txt)",
            options=QFileDialog.DontUseNativeDialog
        )
        if filename != "":
            points = API.import_wire(filename)
            self.gui.sidebar_left.wire_widget.set_wire(
                _points_=points, _stretch_=DefaultStretch, _rotational_symmetry_=DefaultRotationalSymmetry
            )

    def wire_export(self) -> None:
        """
        Exports wire points to a TXT file.
        """
        Debug(self.gui, ".wire_export()")

        if not self.gui.model.wire.valid:
            Assert_Dialog(False, "Attempting to export invalid wire")
            return

        action = QSaveAction(
            self.gui,
            title="Export Wire",
            date=True,
            filename="MagnetiCalc_Wire",
            extension=".txt",
            _filter="Text File (*.txt)"
        )
        if action.filename:
            API.export_wire(action.filename, self.gui.model.wire.points_sliced)



""" Wire_Presets module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Dict
import numpy as np
from magneticalc.Assert_Dialog import Assert_Dialog


class Wire_Presets:
    """ Wire_Presets class. """

    # Preset: A straight line.
    StraightLine = {
        "id": "Straight Line",
        "points": [
            [-1/2, 0, 0],
            [+1/2, 0, 0]
        ]
    }

    # Preset: A single square loop with offset connections.
    SingleSquareLoop_Offset = {
        "id": "Single Square Loop (offset)",
        "points": [
            [-1/2, +1/2, +1/2],
            [+0/2, +1/2, +1/2],
            [+0/2, -1/2, +1/2],
            [+0/2, -1/2, -1/2],
            [+0/2, +1/2, -1/2],
            [+0/2, +1/2, +1/2],
            [+1/2, +1/2, +1/2]
        ]
    }

    # Preset: A single square loop with centered connections.
    SingleSquareLoop_Centered = {
        "id": "Single Square Loop (centered)",
        "points": [
            [-1/2, +1/2, +0/2],
            [+0/2, +1/2, +0/2],
            [+0/2, +1/2, +1/2],
            [+0/2, -1/2, +1/2],
            [+0/2, -1/2, -1/2],
            [+0/2, +1/2, -1/2],
            [+0/2, +1/2, +0/2],
            [+1/2, +1/2, +0/2]
        ]
    }

    # Preset: A "compensated" double square loop with offset connections.
    CompensatedDoubleSquareLoop_Offset = {
        "id": "Compensated Double Square Loop (offset)",
        "points": [
            [-3/6, +1/2, +1/2],
            [-1/6, +1/2, +1/2],
            [-1/6, -1/2, +1/2],
            [-1/6, -1/2, -1/2],
            [-1/6, +1/2, -1/2],
            [-1/6, +1/2, +1/2],
            [+1/6, +1/2, +1/2],
            [+1/6, +1/2, -1/2],
            [+1/6, -1/2, -1/2],
            [+1/6, -1/2, +1/2],
            [+1/6, +1/2, +1/2],
            [+3/6, +1/2, +1/2]
        ]
    }

    # Preset: A "compensated" double square loop with centered connections.
    CompensatedDoubleSquareLoop_Centered = {
        "id": "Compensated Double Square Loop (centered)",
        "points": [
            [-3/6, +1/2, +0/2],
            [-1/6, +1/2, +0/2],
            [-1/6, +1/2, +1/2],
            [-1/6, -1/2, +1/2],
            [-1/6, -1/2, -1/2],
            [-1/6, +1/2, -1/2],
            [-1/6, +1/2, +0/2],
            [+1/6, +1/2, +0/2],
            [+1/6, +1/2, -1/2],
            [+1/6, -1/2, -1/2],
            [+1/6, -1/2, +1/2],
            [+1/6, +1/2, +1/2],
            [+1/6, +1/2, +0/2],
            [+3/6, +1/2, +0/2]
        ]
    }

    # Preset: A single circular loop with offset connections.
    SingleCircularLoop_Offset = {
        "id": "Single Circular Loop (offset)",
        "points": [
            [
                0,
                -np.cos(2 * np.pi * i / 32) / 2,
                +np.sin(2 * np.pi * i / 32) / 2
            ]
            for i in range(32)
        ]
    }

    # Preset: A solenoid of 4 circular loops.
    SolenoidCircularLoops4 = {
        "id": "Solenoid: 4 circular loops",
        "points": [
            [
                i / 129 - 1 / 2,
                -np.cos(2 * np.pi * i / 32) / 2,
                +np.sin(2 * np.pi * i / 32) / 2
            ]
            for i in range(129)
        ]
    }

    # Preset: A solenoid of 8 circular loops.
    SolenoidCircularLoops8 = {
        "id": "Solenoid: 8 circular loops",
        "points": [
            [
                i / 257 - 1 / 2,
                -np.cos(2 * np.pi * i / 32) / 2,
                +np.sin(2 * np.pi * i / 32) / 2
            ]
            for i in range(257)
        ]
    }

    # Preset: A compensated solenoid of 2x 4 circular loops.
    CompensatedSolenoidCircularLoops4 = {
        "id": "Compensated Solenoid: 2x 4 circular loops",
        "points": [
            [
                i / 256 - 1/2,
                -np.cos(+2 * np.pi * i / 32) / 2,
                +np.sin(+2 * np.pi * i / 32) / 2
            ]
            for i in range(128)
        ] +
        [
            [
                i / 256,
                -np.cos(-2 * np.pi * (i + 1) / 32) / 2,
                +np.sin(-2 * np.pi * (i + 1) / 32) / 2
            ]
            for i in range(128)
        ]
    }

    # Preset: A compensated solenoid of 2x 8 circular loops.
    CompensatedSolenoidCircularLoops8 = {
        "id": "Compensated Solenoid: 2x 8 circular loops",
        "points": [
            [
                i / 512 - 1/2,
                -np.cos(+2 * np.pi * i / 32) / 2,
                +np.sin(+2 * np.pi * i / 32) / 2
            ]
            for i in range(256)
        ] +
        [
            [
                i / 512,
                -np.cos(-2 * np.pi * (i + 1) / 32) / 2,
                +np.sin(-2 * np.pi * (i + 1) / 32) / 2
            ]
            for i in range(256)
        ]
    }

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    @staticmethod
    def get_phase_jumping_toroidal_loop(
            n_points: int, n_phase_jumps: int,
            toroidal_radius: float, poloidal_radius: float,
            toroidal_freq: float, poloidal_freq: float,
            flat: bool = True,
    ) -> np.ndarray:
        """
        Generates a (phase-jumping) toroidal loop.

        @param n_points: Number of points
        @param n_phase_jumps: Number of phase jumps
        @param toroidal_radius: Toroidal radius
        @param poloidal_radius: Poloidal radius
        @param toroidal_freq: Toroidal frequency
        @param poloidal_freq: Poloidal frequency
        @param flat: Flat mode
        """
        def rotate_xy(V: np.ndarray, A: float) -> np.ndarray:
            """
            Rotates a 3D vector some angle in the XY-plane.

            @param V: 3D vector
            @param A: Angle in radians
            """
            return np.array(
                [V[0] * np.cos(A) - V[1] * np.sin(A), V[0] * np.sin(A) + V[1] * np.cos(A), V[2]]
            )

        wire_points = []
        for t in range(n_points + 1):
            phase_toroidal = 2 * np.pi * t / n_points * toroidal_freq
            phase_jump_poloidal = (1 if (t // (n_points / n_phase_jumps)) % 2 == 0 else -1)
            phase_poloidal = 2 * np.pi * t / n_points * poloidal_freq * phase_jump_poloidal + (0 if flat else np.pi / 2)
            phase_poloidal = np.fmod(phase_poloidal, 2 * np.pi)
            p_toroidal = toroidal_radius * np.array([np.cos(phase_toroidal), np.sin(phase_toroidal), 0])
            p_poloidal = poloidal_radius * np.array([0, np.cos(phase_poloidal), np.sin(phase_poloidal)])
            wire_points.append(p_toroidal + rotate_xy(p_poloidal, phase_toroidal + (0 if flat else np.pi / 2)))
        return np.array(wire_points)

    # Preset: A phase-jumping toroidal loop: 8 turns.
    PhaseJumpingToroidalLoop8 = {
        "id": "Phase-jumping Toroidal Loop: 8 turns",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=640, n_phase_jumps=8,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=8
        )
    }
    # Preset: A phase-jumping toroidal loop: 32 turns.
    PhaseJumpingToroidalLoop16 = {
        "id": "Phase-jumping Toroidal Loop: 16 turns",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=640, n_phase_jumps=16,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=16
        )
    }
    # Preset: A phase-jumping toroidal loop: 32 turns, hi-res.
    PhaseJumpingToroidalLoop32HiRes = {
        "id": "Phase-jumping Toroidal Loop: 32 turns, hi-res",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=2000, n_phase_jumps=32,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=32
        )
    }
    # Preset: A phase-jumping toroidal loop: 32 turns, lo-res.
    PhaseJumpingToroidalLoop32LoRes = {
        "id": "Phase-jumping Toroidal Loop: 32 turns, lo-res",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=640, n_phase_jumps=32,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=32
        )
    }
    # Preset: A phase-jumping toroidal loop: 32 turns, hi-res, flat.
    PhaseJumpingToroidalLoop32HiResFlat = {
        "id": "Phase-jumping Toroidal Loop: 32 turns, hi-res, flat",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=2000, n_phase_jumps=32,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=32,
            flat=True,
        )
    }
    # Preset: A phase-jumping toroidal loop: 32 turns, lo-res, flat.
    PhaseJumpingToroidalLoop32LoResFlat = {
        "id": "Phase-jumping Toroidal Loop: 32 turns, lo-res, flat",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=640, n_phase_jumps=32,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=32,
            flat=True,
        )
    }

    # Preset: A toroidal loop: 8 turns.
    ToroidalLoop8 = {
        "id": "Toroidal Loop: 8 turns",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=640, n_phase_jumps=1,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=8
        )
    }

    # Preset: A toroidal loop: 16 turns.
    ToroidalLoop16 = {
        "id": "Toroidal Loop: 16 turns",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=640, n_phase_jumps=1,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=16
        )
    }

    # Preset: A toroidal loop: 32 turns, hi-res.
    ToroidalLoop32HiRes = {
        "id": "Toroidal Loop: 32 turns, hi-res",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=2000, n_phase_jumps=1,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=32,
            flat=False,
        )
    }
    ToroidalLoop32LoRes = {
        "id": "Toroidal Loop: 32 turns, lo-res",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=640, n_phase_jumps=1,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=32,
            flat=False,
        )
    }

    # Preset: A toroidal loop: 32 turns, hi-res, flat.
    ToroidalLoop32HiResFlat = {
        "id": "Toroidal Loop: 32 turns, hi-res, flat",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=2000, n_phase_jumps=1,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=32,
            flat=True,
        )
    }
    ToroidalLoop32LoResFlat = {
        "id": "Toroidal Loop: 32 turns, lo-res, flat",
        "points": get_phase_jumping_toroidal_loop.__func__(  # type: ignore
            n_points=640, n_phase_jumps=1,
            toroidal_radius=1, poloidal_radius=.5,
            toroidal_freq=1, poloidal_freq=32,
            flat=True,
        )
    }

    # ------------------------------------------------------------------------------------------------------------------

    # List of all above presets
    List = [
        StraightLine,
        SingleSquareLoop_Offset,
        SingleSquareLoop_Centered,
        CompensatedDoubleSquareLoop_Offset,
        CompensatedDoubleSquareLoop_Centered,
        SingleCircularLoop_Offset,
        SolenoidCircularLoops4,
        SolenoidCircularLoops8,
        CompensatedSolenoidCircularLoops4,
        CompensatedSolenoidCircularLoops8,
        PhaseJumpingToroidalLoop8,
        PhaseJumpingToroidalLoop16,
        PhaseJumpingToroidalLoop32HiRes,
        PhaseJumpingToroidalLoop32LoRes,
        PhaseJumpingToroidalLoop32HiResFlat,
        PhaseJumpingToroidalLoop32LoResFlat,
        ToroidalLoop8,
        ToroidalLoop16,
        ToroidalLoop32HiRes,
        ToroidalLoop32LoRes,
        ToroidalLoop32HiResFlat,
        ToroidalLoop32LoResFlat,
    ]

    # ------------------------------------------------------------------------------------------------------------------

    Fallback = StraightLine

    @staticmethod
    def get_by_id(_id_: str) -> Dict:
        """
        Selects a preset by name.

        @param _id_: Preset ID
        @return: Preset parameters
        """
        for preset in Wire_Presets.List:
            if _id_ == preset["id"]:
                return preset

        Assert_Dialog(False, f"Invalid wire preset ID: Defaulting to \"{Wire_Presets.Fallback['id']}\"")
        return Wire_Presets.Fallback



""" QSliderFloat module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QSlider


class QSliderFloat(QSlider):
    """ QSliderFloat class. """

    def __init__(
            self,
            orientation: Qt.Orientation,
            minimum: float,
            maximum: float,
            step: float
    ) -> None:
        """
        Initializes a slider supporting float values.

        @param orientation: Orientation
        @param step: Step value
        @param minimum: Minimum value
        @param maximum: Maximum value
        """
        QSlider.__init__(self, orientation)
        self._step = step
        self.setMinimum(round(minimum / self._step))
        self.setMaximum(round(maximum / self._step))
        self.setSingleStep(1)

    def setValue(self, value: float) -> None:
        """
        Sets the slider value.

        @param value: Slider value
        """
        super().setValue(round(value / self._step))

    def get_value(self) -> float:
        """
        Returns the slider value.

        @return: Slider value
        """
        return super().value() * self._step



""" SamplingVolume_Widget module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Optional, List, Tuple
import numpy as np
from PyQt5.QtWidgets import QHBoxLayout, QComboBox, QWidget, QSizePolicy
from magneticalc.QtWidgets2.QGroupBox2 import QGroupBox2
from magneticalc.QtWidgets2.QHLine import QHLine
from magneticalc.QtWidgets2.QIconLabel import QIconLabel
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.QtWidgets2.QPushButton2 import QPushButton2
from magneticalc.QtWidgets2.QSpinBox2 import QSpinBox2
from magneticalc.Constraint_Editor import Constraint_Editor
from magneticalc.Debug import Debug
from magneticalc.ModelAccess import ModelAccess
from magneticalc.OverridePadding_Dialog import OverridePadding_Dialog
from magneticalc.Theme import Theme


class SamplingVolume_Widget(QGroupBox2):
    """ SamplingVolume_Widget class. """

    # Display settings
    UnitsLabelWidth = 26

    # Spinbox limits
    PaddingMin = -1000
    PaddingMax = +1000

    # Resolution options
    ResolutionOptionsDict = {
        "1024"       : 10,
        "512"       : 9,
        "256"       : 8,
        "128"       : 7,
        "64"        : 6,
        "32"        : 5,
        "16"        : 4,
        "8"         : 3,
        "4"         : 2,
        "2"         : 1,
        "1"         : 0,
        "1 / 2"     : -1,
        "1 / 4"     : -2,
        "1 / 8"     : -3,
        "1 / 16"    : -4,
        "1 / 32"    : -5,
        "1 / 64"    : -6,
        "1 / 128"   : -7,
        "1 / 256"   : -8,
        "1 / 512"   : -9,
        "1 / 1024"   : -10,
    }

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Populates the widget.

        @param gui: GUI
        """
        QGroupBox2.__init__(self, "Sampling Volume", color=Theme.DarkColor)
        Debug(self, ": Init", init=True)
        self.gui = gui

        # Insert constraint editor, but don't fully initialize it yet
        self.constraint_editor = Constraint_Editor(self.gui)

        padding_icon_label = QIconLabel("Padding", "mdi.arrow-expand-all", color=Theme.DarkColor)
        padding_override_button = QPushButton2("  Override …  ", "", self.override_padding)
        padding_icon_label.addWidget(padding_override_button)
        padding_icon_label.addWidget(QPushButton2("", "fa.eraser", self.clear_padding))
        padding_icon_label.addWidget(QLabel2("cm", align_right=True, width=self.UnitsLabelWidth))
        self.addLayout(padding_icon_label)

        self.padding_widget = QWidget()
        padding_layout = QHBoxLayout()
        self.padding_spinbox = []
        for i in range(3):
            padding_spinbox = QSpinBox2(
                minimum=self.PaddingMin,
                maximum=self.PaddingMax,
                value=0,
                value_changed=self.update_padding
            )
            self.padding_spinbox.append(padding_spinbox)
            padding_layout.addWidget(QLabel2(["X", "Y", "Z"][i] + ":", expand=False))
            padding_layout.addWidget(self.padding_spinbox[i])
        self.padding_widget.setLayout(padding_layout)
        self.addWidget(self.padding_widget)

        total_extent_layout = QHBoxLayout()
        self.total_extent_label = QLabel2("N/A", color=Theme.MainColor, align_right=True)
        total_extent_layout.addWidget(QLabel2("Total extent:", italic=True, color=Theme.LiteColor))
        total_extent_layout.addWidget(self.total_extent_label)
        self.addLayout(total_extent_layout)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addWidget(QHLine())

        constraints_icon_label = QIconLabel("Constraints", "mdi.playlist-edit", color=Theme.DarkColor)
        constraints_icon_label.addWidget(QLabel2("⟨F3⟩", font_size="13px", color=Theme.LiteColor, expand=False))
        constraints_icon_label.addWidget(QPushButton2("Edit …", "", self.open_constraint_editor))
        self.addLayout(constraints_icon_label)

        total_constraints_layout = QHBoxLayout()
        self.total_constraints_label = QLabel2("N/A", color=Theme.MainColor, align_right=True)
        total_constraints_layout.addWidget(QLabel2("Total constraints:", italic=True, color=Theme.LiteColor))
        total_constraints_layout.addWidget(self.total_constraints_label)
        self.addLayout(total_constraints_layout)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.addWidget(QHLine())

        self.addLayout(QIconLabel("Resolution", "fa.th", color=Theme.DarkColor))

        self.resolution_combobox = QComboBox()
        self.resolution_combobox.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        resolution_layout = QHBoxLayout()
        resolution_layout.addWidget(self.resolution_combobox)
        self.resolution_units_label = QLabel2(" Points / cm", expand=False)
        resolution_layout.addWidget(self.resolution_units_label)
        self.addLayout(resolution_layout)

        # Populate resolution combobox
        for i, value in enumerate(self.ResolutionOptionsDict):
            self.resolution_combobox.addItem(str(value))
        self.resolution_combobox.currentIndexChanged.connect(  # type: ignore
            lambda: self.set_sampling_volume(
                _resolution_exponent_=self.ResolutionOptionsDict.get(
                    self.resolution_combobox.currentText(),
                    0
                )
            )
        )

        total_points_layout = QHBoxLayout()
        self.total_points_label = QLabel2("N/A", color=Theme.MainColor, align_right=True)
        total_points_layout.addWidget(QLabel2("Total sampling points:", italic=True, color=Theme.LiteColor))
        total_points_layout.addWidget(self.total_points_label)
        self.addLayout(total_points_layout)

    def reload(self) -> None:
        """
        Reloads the widget.
        """
        Debug(self, ".reload()", refresh=True)

        self.blockSignals(True)

        sampling_volume_padding = self.gui.project.get_point("sampling_volume_padding")
        for i in range(3):
            self.padding_spinbox[i].setValue(int(sampling_volume_padding[i]))

        # Set default resolution if it is not available anymore
        target = self.gui.project.get_int("sampling_volume_resolution_exponent")
        if target not in self.ResolutionOptionsDict.values():
            Debug(self, f": WARNING: Invalid: sampling_volume_resolution_exponent = {target}", warning=True)
            self.gui.project.set_int("sampling_volume_resolution_exponent", 0)

        # Select the resolution
        target = self.gui.project.get_int("sampling_volume_resolution_exponent")
        for i, value in enumerate(self.ResolutionOptionsDict.values()):
            if value == target:
                self.resolution_combobox.setCurrentIndex(i)

        self.blockSignals(False)

        # Reload the constraint editor
        self.constraint_editor.reload()

        # Initially load sampling volume from project
        self.set_sampling_volume(recalculate=False, invalidate=False)

        self.update()

    def update(self) -> None:
        """
        Updates the widget.
        """
        Debug(self, ".update()", refresh=True)

        self.update_labels()
        self.update_controls()

    def update_labels(self) -> None:
        """
        Updates the labels.
        """
        Debug(self, ".update_labels()", refresh=True)

        if self.gui.model.sampling_volume.valid:
            self.total_extent_label.setText(
                " × ".join([f"{extent:.0f}" for extent in self.gui.model.sampling_volume.extent]) + " cm³"
            )
            self.total_points_label.setText(
                str(self.gui.model.sampling_volume.points_count)
            )
        else:
            self.total_extent_label.setText("N/A")
            self.total_points_label.setText("N/A")

        self.total_constraints_label.setText(str(self.constraint_editor.rows_count))

    def update_controls(self) -> None:
        """
        Updates the controls.
        """
        Debug(self, ".update_controls()", refresh=True)

        self.padding_widget.setEnabled(not self.gui.project.get_bool("sampling_volume_override_padding"))

        self.set_color(Theme.MainColor if self.gui.model.sampling_volume.valid else Theme.FailureColor)

    # ------------------------------------------------------------------------------------------------------------------

    def update_padding(self) -> None:
        """
        Updates padding.
        """
        if self.signalsBlocked():
            return

        Debug(self, ".update_padding()")

        self.set_sampling_volume(_padding_=[self.padding_spinbox[i].value() for i in range(3)])

    def clear_padding(self) -> None:
        """
        Clears the padding values.
        """
        Debug(self, ".clear_padding()")

        self.gui.project.set_bool("sampling_volume_override_padding", False)
        self.gui.project.set_points("sampling_volume_bounding_box", [np.zeros(3), np.zeros(3)])

        self.set_sampling_volume(_padding_=[0, 0, 0])

        self.blockSignals(True)
        for i in range(3):
            self.padding_spinbox[i].setValue(0)
        self.blockSignals(False)

    def override_padding(self) -> None:
        """
        Override padding, setting the bounding box directly.
        """
        Debug(self, ".override_padding()")

        dialog = OverridePadding_Dialog(self.gui)
        dialog.show()

        if not dialog.user_accepted:
            return

        # This needlessly also clears the override padding and bounding box settings, but that's acceptable.
        self.clear_padding()

        self.set_sampling_volume(
            _override_padding_=True,
            _bounding_box_=(
                np.array([dialog.bounds_min_spinbox[i].value() for i in range(3)]),
                np.array([dialog.bounds_max_spinbox[i].value() for i in range(3)])
            )
        )

        self.update_controls()

    # ------------------------------------------------------------------------------------------------------------------

    def open_constraint_editor(self) -> None:
        """
        Opens the constraint editor.
        Recalculates the sampling volume if the constraints changed.
        """
        Debug(self, ".open_constraint_editor()")

        self.constraint_editor.show()

        if self.constraint_editor.changed:
            self.set_sampling_volume()
            self.constraint_editor.changed = False

    # ------------------------------------------------------------------------------------------------------------------

    def set_sampling_volume(
            self,
            _resolution_exponent_: Optional[int] = None,
            _label_resolution_exponent_: Optional[int] = None,
            _padding_: Optional[List] = None,
            _override_padding_: Optional[bool] = None,
            _bounding_box_: Optional[Tuple[np.ndarray, np.ndarray]] = None,
            invalidate: bool = True,
            recalculate: bool = True
    ) -> None:
        """
        Sets the sampling volume. This will overwrite the currently set sampling volume in the model.
        Any underscored parameter may be left set to None in order to load its default value.

        @param _resolution_exponent_: Sampling volume resolution exponent
        @param _label_resolution_exponent_: Sampling volume label resolution exponent
        @param _padding_: Padding (3D point)
        @param _override_padding_: Enable to override padding instead, setting the bounding box directly
        @param _bounding_box_: Bounding box (used in conjunction with override_padding)
        @param invalidate: Enable to invalidate this model hierarchy level
        @param recalculate: Enable to trigger final recalculation
        """
        if self.signalsBlocked():
            return

        Debug(self, ".set_sampling_volume()")

        with ModelAccess(self.gui, recalculate):

            resolution_exponent = self.gui.project.set_get_int(
                "sampling_volume_resolution_exponent",
                _resolution_exponent_
            )

            label_resolution_exponent = self.gui.project.set_get_int(
                "sampling_volume_label_resolution_exponent",
                _label_resolution_exponent_
            )

            resolution = np.power(2.0, resolution_exponent)
            label_resolution = np.power(2.0, label_resolution_exponent)

            # Convert every line in the constraint editor to a Constraint() object
            self.gui.model.set_sampling_volume(
                invalidate=invalidate,
                resolution=resolution,
                label_resolution=label_resolution,
                constraints=self.constraint_editor.get_constraints()
            )

            self.readjust(_padding_=_padding_, _override_padding_=_override_padding_, _bounding_box_=_bounding_box_)

            if recalculate:
                # The display widget depends on the sampling volume
                self.gui.sidebar_right.display_widget.update()

                # This forces updating the number of constraints if the sampling volume is already invalidated
                if not self.gui.model.sampling_volume.valid:
                    self.update()

    # ------------------------------------------------------------------------------------------------------------------

    def readjust(
            self,
            _padding_: Optional[List] = None,
            _override_padding_: Optional[bool] = None,
            _bounding_box_: Optional[Tuple[np.ndarray, np.ndarray]] = None,
    ) -> None:
        """
        Readjusts the sampling volume to the currently set wire bounds and padding.
        This also readjusts the minimum padding bounds in case the wire bounds have shrunk too far.

        Padding may also be overridden instead, setting the bounding box directly.

        Any underscored parameter may be left set to None in order to load its default value.

        @param _padding_: Padding (3D point)
        @param _override_padding_: Enable to override padding instead, setting the bounding box directly
        @param _bounding_box_: Bounding box (used in conjunction with override_padding)
        """
        Debug(self, ".readjust()")

        override_padding = self.gui.project.set_get_bool("sampling_volume_override_padding", _override_padding_)
        bounding_box = self.gui.project.set_get_points("sampling_volume_bounding_box", _bounding_box_)

        self.gui.model.sampling_volume.set_bounds_nearest(
            bounding_box if override_padding else self.gui.model.wire.bounds
        )

        if not override_padding:
            bounds = self.gui.model.sampling_volume.bounds
            padding = self.gui.project.set_get_point("sampling_volume_padding", _padding_)

            # Adjust padding values if they are now outside the new minimum padding bounds
            padding_exceeded = [bounds[0][i] > self.padding_spinbox[i].value() for i in range(3)]
            padding = [bounds[0][i] if padding_exceeded[i] else padding[i] for i in range(3)]
            if any(padding_exceeded):
                Debug(self, ".readjust(): Readjusted bounding box")
                self.gui.project.set_point("sampling_volume_padding", padding)

            self.blockSignals(True)
            [self.padding_spinbox[i].setMinimum(int(bounds[0][i])) for i in range(3)]
            self.blockSignals(False)

            self.gui.model.sampling_volume.set_padding_nearest(padding)



""" Comparison_Types module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from magneticalc.Assert_Dialog import Assert_Dialog


""" Comparison type: In Range. """
COMPARISON_TYPE_IN_RANGE = 0

""" Comparison type: Not In Range. """
COMPARISON_TYPE_NOT_IN_RANGE = 1


""" Map of comparison types to names. """
Comparison_Types_Names_Map = {
    COMPARISON_TYPE_IN_RANGE        : "In Range",
    COMPARISON_TYPE_NOT_IN_RANGE    : "Not In Range",
}

""" Default comparison type. """
Comparison_Type_Default = COMPARISON_TYPE_IN_RANGE


def comparison_type_safe(comparison_type: int) -> int:
    """
    A valid comparison type is passed through, but an invalid comparison type converts to the default type.

    @param comparison_type: Comparison type
    @return: Safe comparison type
    """
    if comparison_type in Comparison_Types_Names_Map:
        return comparison_type
    else:
        Assert_Dialog(False, "Invalid comparison type, using fallback type")
        return Comparison_Type_Default


def comparison_type_to_name(comparison_type: int) -> str:
    """
    Converts a comparison type to a comparison name.

    @param comparison_type: Comparison type
    @return: Comparison name
    """
    return Comparison_Types_Names_Map[comparison_type_safe(comparison_type)]


def comparison_name_to_type(comparison_name: str) -> int:
    """
    Converts a comparison name to a comparison type.

    @param comparison_name: Comparison name
    @return Comparison type
    """
    return {_name: _type for _type, _name in Comparison_Types_Names_Map.items()}[comparison_name]
